from pathlib import Path
from typing import Annotated, Any

import typer
from rich import print

from cognite_toolkit._cdf_tk.commands import (
    ProfileAssetCentricCommand,
    ProfileAssetCommand,
    ProfileRawCommand,
    ProfileTransformationCommand,
)
from cognite_toolkit._cdf_tk.utils.auth import EnvironmentVariables


class ProfileApp(typer.Typer):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.callback(invoke_without_command=True)(self.main)
        self.command("assets")(self.assets)
        self.command("asset-centric")(self.asset_centric)
        self.command("transformations")(self.transformations)
        self.command("raw")(self.raw)

    def main(self, ctx: typer.Context) -> None:
        """Commands profile functionality"""
        if ctx.invoked_subcommand is None:
            print("Use [bold yellow]cdf profile --help[/] for more information.")

    @staticmethod
    def assets(
        ctx: typer.Context,
        hierarchy: Annotated[
            str | None,
            typer.Option(
                "--hierarchy",
                "-h",
                help="The asset hierarchy to profile. This should be the externalId of the root asset. If not provided,"
                " an interactive prompt will be used to select the hierarchy.",
            ),
        ] = None,
        output_spreadsheet: Annotated[
            Path | None,
            typer.Option(
                "--output-spreadsheet",
                "-o",
                help="The path to the output spreadsheet. If not provided, the output will only be printed to the console.",
            ),
        ] = None,
        verbose: bool = False,
    ) -> None:
        """This command gives an overview over the assets in the given hierarchy.
        It works by listing all assets, events, files, timeseries, and sequences related to the given hierarchy.
        In addition, it lists the data sets that is used for each of the resources, the transformations that writes to
        these data sets, and the RAW tables that is used in these transformations.
        """
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = ProfileAssetCommand(output_spreadsheet)
        cmd.run(
            lambda: cmd.assets(
                client=client,
                hierarchy=hierarchy,
                verbose=verbose,
            )
        )

    @staticmethod
    def asset_centric(
        ctx: typer.Context,
        hierarchy: Annotated[
            str | None,
            typer.Option(
                "--hierarchy",
                "-h",
                help="The asset hierarchy to profile. This should be the externalId of the root asset. If not provided,"
                " an interactive prompt will be used to select the hierarchy (or select all assets-centric assets).",
            ),
        ] = None,
        output_spreadsheet: Annotated[
            Path | None,
            typer.Option(
                "--output-spreadsheet",
                "-o",
                help="The path to the output spreadsheet. If not provided, the output will only be printed to the console.",
            ),
        ] = None,
        verbose: bool = False,
    ) -> None:
        """This command gives an overview over the metadata and labels for each of the asset-centric resources.
        This shows an approximation of unstructured data count. This can, for example, be used to estimate the
        effort to model this data in data modeling."""
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = ProfileAssetCentricCommand(output_spreadsheet)
        cmd.run(
            lambda: cmd.asset_centric(
                client,
                hierarchy,
                verbose,
            )
        )

    @staticmethod
    def transformations(
        destination: Annotated[
            str | None,
            typer.Option(
                "--destination",
                "-d",
                help="Destination type the transformations data should be written to. This can be 'assets', 'events', 'files',"
                "'timeseries', or 'sequences'. If not provided, and interactive mode is enabled, the user will be prompted to select a destination.",
            ),
        ] = None,
        output_spreadsheet: Annotated[
            Path | None,
            typer.Option(
                "--output-spreadsheet",
                "-o",
                help="The path to the output spreadsheet. If not provided, the output will only be printed to the console.",
            ),
        ] = None,
        verbose: bool = False,
    ) -> None:
        """This command gives an overview over the transformations that write to the given destination.
        It works by checking all transformations that writes to the given destination, lists the sources of the data,
        and the target columns.
        This is intended to show the flow of data from raw into CDF. This can, for example, be used to determine the
        source of the data in a specific CDF resource.
        """
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = ProfileTransformationCommand(output_spreadsheet)
        cmd.run(
            lambda: cmd.transformation(
                client,
                destination,
                verbose,
            )
        )

    @staticmethod
    def raw(
        ctx: typer.Context,
        destination: Annotated[
            str | None,
            typer.Option(
                "--destination",
                "-d",
                help="Destination type the raw data should be written to. This can be 'assets', 'events', 'files',"
                "'timeseries', or 'sequences'. If not provided, and interactive mode is enabled, the user will be prompted to select a destination.",
            ),
        ] = None,
        output_spreadsheet: Annotated[
            Path | None,
            typer.Option(
                "--output-spreadsheet",
                "-o",
                help="The path to the output spreadsheet. If not provided, the output will only be printed to the console.",
            ),
        ] = None,
        verbose: bool = False,
    ) -> None:
        """This command gives an overview over the staging tables in CDF and where they are used.
        It works by checking all transformations that writes to the given destination and lists all the raw tables
        that are used in those transformations.
        This is intended to show the flow of data from raw into CDF. This can, for example, be used to determine the
        source of the data in a specific CDF resource.
        """
        client = EnvironmentVariables.create_from_environment().get_client()
        cmd = ProfileRawCommand(output_spreadsheet)
        cmd.run(
            lambda: cmd.raw(
                client,
                destination,
                verbose,
            )
        )



from pathlib import Path

import pytest
from cognite.client.data_classes.data_modeling import NodeId, ViewId

from cognite_toolkit._cdf_tk.commands._migrate.data_classes import (
    MigrationMappingList,
    TimeSeriesMapping,
    TimeSeriesMigrationMappingList,
)


class TestMigrationMappingList:
    @pytest.mark.parametrize(
        "content, expected",
        [
            pytest.param(
                "id,dataSetId,space,externalId\n123,123,sp_full_ts,full_ts_id\n3231,,sp_step_ts,step_ts_id\n",
                MigrationMappingList(
                    [
                        TimeSeriesMapping(
                            id=123,
                            dataSetId=123,
                            instanceId=NodeId("sp_full_ts", "full_ts_id"),
                        ),
                        TimeSeriesMapping(
                            id=3231,
                            dataSetId=None,
                            instanceId=NodeId("sp_step_ts", "step_ts_id"),
                        ),
                    ]
                ),
                id="Mapping IDs with dataSetId",
            ),
            pytest.param(
                "id,space,externalId\n230,my_space,target_external_id\n",
                MigrationMappingList(
                    [
                        TimeSeriesMapping(
                            id=230,
                            dataSetId=None,
                            instanceId=NodeId("my_space", "target_external_id"),
                        )
                    ]
                ),
                id="Mapping IDs without dataSetId",
            ),
            pytest.param(
                """\ufeffid,dataSetId,space,externalId\n42,123,sp_full_ts,full_ts_id\n""",
                MigrationMappingList(
                    [
                        TimeSeriesMapping(
                            id=42,
                            dataSetId=123,
                            instanceId=NodeId("sp_full_ts", "full_ts_id"),
                        )
                    ]
                ),
                id="Mapping with BOM",
            ),
            pytest.param(
                """id,space,externalId,dataSetId,ingestionView,consumerViewSpace,consumerViewExternalId,consumerViewVersion\n
123,sp_full_ts,full_ts_id,123,ingestion_view_id,consumer_view_space,consumer_view_external_id,1.0\n
3231,sp_step_ts,step_ts_id,,ingestion_view_id_2,consumer_view_space_2,consumer_view_external_id_2,2.0\n""",
                MigrationMappingList(
                    [
                        TimeSeriesMapping(
                            id=123,
                            dataSetId=123,
                            instanceId=NodeId("sp_full_ts", "full_ts_id"),
                            ingestionView="ingestion_view_id",
                            preferredConsumerView=ViewId(
                                space="consumer_view_space", external_id="consumer_view_external_id", version="1.0"
                            ),
                        ),
                        TimeSeriesMapping(
                            id=3231,
                            dataSetId=None,
                            instanceId=NodeId("sp_step_ts", "step_ts_id"),
                            ingestionView="ingestion_view_id_2",
                            preferredConsumerView=ViewId(
                                space="consumer_view_space_2", external_id="consumer_view_external_id_2", version="2.0"
                            ),
                        ),
                    ]
                ),
                id="Mapping with all columns including optional ones",
            ),
        ],
    )
    def test_read_mapping_file(self, content: str, expected: MigrationMappingList, tmp_path: Path) -> None:
        input_file = tmp_path / "mapping_file.csv"
        input_file.write_text(content, encoding="utf-8")
        actual = TimeSeriesMigrationMappingList.read_csv_file(input_file)
        assert not actual.invalid_rows
        assert actual == expected

    @pytest.mark.parametrize(
        "content, expected_msg",
        [
            pytest.param("", "No data found in the file: '{filepath}'.", id="empty_file"),
            pytest.param(
                "id,externalId\n123,full_ts_id\n",
                "Missing required columns: space",
                id="Too few columns",
            ),
        ],
    )
    def test_read_invalid_mapping_file(self, content: str, expected_msg: str, tmp_path: Path) -> None:
        input_file = tmp_path / "mapping_file.csv"
        input_file.write_text(content, encoding="utf-8")
        if "{filepath}" in expected_msg:
            expected_msg = expected_msg.format(filepath=input_file.as_posix())

        with pytest.raises(ValueError) as exc_info:
            TimeSeriesMigrationMappingList.read_csv_file(input_file)
        assert str(exc_info.value) == expected_msg

    @pytest.mark.parametrize(
        "content, unexpected_columns",
        [
            pytest.param(
                "id,dataSetId,space,externalId,unexpected_col\n123,123,sp_full_ts,full_ts_id,some_value\n",
                {"unexpected_col"},
                id="Unexpected column",
            ),
        ],
    )
    def test_read_with_warnings(self, content: str, unexpected_columns: list[str], tmp_path: Path) -> None:
        input_file = tmp_path / "mapping_file.csv"
        input_file.write_text(content, encoding="utf-8")

        mapping = TimeSeriesMigrationMappingList.read_csv_file(input_file)
        assert mapping.unexpected_columns == unexpected_columns



import sys
from abc import ABC
from collections.abc import Sequence
from datetime import datetime, timezone
from typing import Any
from uuid import uuid4

from cognite.client import CogniteClient
from cognite.client.data_classes._base import (
    CogniteResource,
    CogniteResourceList,
    WriteableCogniteResource,
    WriteableCogniteResourceList,
)
from cognite.client.data_classes.data_modeling import (
    DirectRelationReference,
    EdgeId,
    NodeId,
)
from cognite.client.data_classes.data_modeling.ids import ViewId
from cognite.client.data_classes.data_modeling.instances import (
    EdgeApply,
    InstanceApply,
    Node,
    NodeApply,
    NodeListWithCursor,
    PropertyOptions,
    T_Node,
    TypedNode,
    TypedNodeApply,
)

from cognite_toolkit._cdf_tk.client.data_classes.migration import AssetCentricId

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


CANVAS_INSTANCE_SPACE = "IndustrialCanvasInstanceSpace"
SOLUTION_TAG_SPACE = "SolutionTagsInstanceSpace"
CANVAS_SCHEMA_SPACE = "cdf_industrial_canvas"
ANNOTATION_EDGE_TYPE = DirectRelationReference(CANVAS_SCHEMA_SPACE, "referencesCanvasAnnotation")
CONTAINER_REFERENCE_EDGE_TYPE = DirectRelationReference(CANVAS_SCHEMA_SPACE, "referencesContainerReference")
FDM_CONTAINER_REFERENCE_EDGE_TYPE = DirectRelationReference(
    CANVAS_SCHEMA_SPACE, "referencesFdmInstanceContainerReference"
)


class ExtendedTypedNodeApply(TypedNodeApply, ABC):
    def dump(self, camel_case: bool = True, keep_existing_version: bool = True) -> dict[str, Any]:
        """Dumps the instance to a dictionary.

        Args:
            camel_case: If True, the keys will be in camelCase format.
            keep_existing_version: Whether to keep the existing version in the dumped data. Typically,
                you keep the existing version if you want to update the instance, and you remove it if
                you are creating a new instance (for example using the current node as a template).

        Returns:
            A dictionary representation of the instance.

        """
        output = super().dump(camel_case=camel_case)
        if not keep_existing_version:
            output.pop("existingVersion" if camel_case else "existing_version", None)
        return output


class _CanvasProperties:
    created_by = PropertyOptions("createdBy")
    updated_at = PropertyOptions("updatedAt")
    updated_by = PropertyOptions("updatedBy")
    is_locked = PropertyOptions("isLocked")
    source_canvas_id = PropertyOptions("sourceCanvasId")
    is_archived = PropertyOptions("isArchived")
    solution_tags = PropertyOptions("solutionTags")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_industrial_canvas", "Canvas", "v7")


class CanvasApply(_CanvasProperties, ExtendedTypedNodeApply):
    """This represents the writing format of canvas.

    It is used to when data is written to CDF.

    Args:
        external_id: The external id of the canva.
        name: The name or title of the canvas.
        created_by: The user identifier of the user that created the canvas.
        updated_at: The timestamp when the canvas was last updated.
        updated_by: The user identifier of the user that last updated the canvas.
        is_locked: The boolean state for handling canvas locking which is one-way operation from the user perspective
        visibility: The application-level visibility of the canvas. Must be either 'public' or 'private' and, if not
            set, the canvas is expected to be public.
        source_canvas_id: The property for handling versioning. Example sourceCanvasId === selectedCanvas -> query all
            versions of such canvas
        is_archived: Boolean that indicates whether the canvas is archived.
        context: Stores contextual data attached to the canvas, such as rules and pinned values.
        solution_tags: The list of solution tags associated with the canvas.
        existing_version: Fail the ingestion request if the node's version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the node
            (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert,
            so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion
            request, then the item will be skipped instead of failing the ingestion request.
        type: Direct relation pointing to the type node.
    """

    def __init__(
        self,
        external_id: str,
        *,
        name: str,
        created_by: str,
        updated_at: datetime,
        updated_by: str,
        is_locked: bool | None = None,
        visibility: str | None = None,
        source_canvas_id: str | None = None,
        is_archived: bool | None = None,
        context: list[dict] | None = None,
        solution_tags: list[DirectRelationReference | tuple[str, str]] | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        TypedNodeApply.__init__(self, CANVAS_INSTANCE_SPACE, external_id, existing_version, type)
        self.name = name
        self.created_by = created_by
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.is_locked = is_locked
        self.visibility = visibility
        self.source_canvas_id = source_canvas_id
        self.is_archived = is_archived
        self.context = context
        self.solution_tags = (
            [DirectRelationReference.load(solution_tag) for solution_tag in solution_tags] if solution_tags else None
        )


class Canvas(_CanvasProperties, TypedNode):
    """This represents the reading format of canva.

    It is used to when data is read from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the canva.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        name: The name or title of the canvas.
        created_by: The user identifier of the user that created the canvas.
        updated_at: The timestamp when the canvas was last updated.
        updated_by: The user identifier of the user that last updated the canvas.
        is_locked: The boolean state for handling canvas locking which is one-way operation from the user perspective
        visibility: The application-level visibility of the canvas. Must be either 'public' or 'private' and, if not
            set, the canvas is expected to be public.
        source_canvas_id: The property for handling versioning. Example sourceCanvasId === selectedCanvas -> query all
            versions of such canvas
        is_archived: Boolean that indicates whether the canvas is archived.
        context: Stores contextual data attached to the canvas, such as rules and pinned values.
        solution_tags: The list of solution tags associated with the canvas.
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        name: str,
        created_by: str,
        updated_at: datetime,
        updated_by: str,
        is_locked: bool | None = None,
        visibility: str | None = None,
        source_canvas_id: str | None = None,
        is_archived: bool | None = None,
        context: list[dict] | None = None,
        solution_tags: list[DirectRelationReference] | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.name = name
        self.created_by = created_by
        self.updated_at = updated_at
        self.updated_by = updated_by
        self.is_locked = is_locked
        self.visibility = visibility
        self.source_canvas_id = source_canvas_id
        self.is_archived = is_archived
        self.context = context
        self.solution_tags = (
            [DirectRelationReference.load(solution_tag) for solution_tag in solution_tags] if solution_tags else None
        )

    def as_write(self) -> CanvasApply:
        return CanvasApply(
            self.external_id,
            name=self.name,
            created_by=self.created_by,
            updated_at=self.updated_at,
            updated_by=self.updated_by,
            is_locked=self.is_locked,
            visibility=self.visibility,
            source_canvas_id=self.source_canvas_id,
            is_archived=self.is_archived,
            context=self.context,
            solution_tags=self.solution_tags,
            existing_version=self.version,
            type=self.type,
        )


class _CanvasAnnotationProperties:
    id_ = PropertyOptions("id")
    annotation_type = PropertyOptions("annotationType")
    container_id = PropertyOptions("containerId")
    is_selectable = PropertyOptions("isSelectable")
    is_draggable = PropertyOptions("isDraggable")
    is_resizable = PropertyOptions("isResizable")
    properties_ = PropertyOptions("properties")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_industrial_canvas", "CanvasAnnotation", "v1")


class CanvasAnnotationApply(_CanvasAnnotationProperties, ExtendedTypedNodeApply):
    """This represents the writing format of canvas annotation.

    It is used to when data is written to CDF.

    Args:
        external_id: The external id of the canvas annotation.
        id_: The unique identifier of the canvas annotation.
        annotation_type: The type of the annotation. Must be one of rectangle, ellipse, polyline, text or sticky.
        container_id: The optional ID of the container that the annotation is contained in.
        is_selectable: Boolean that indicates whether the annotation is selectable.
        is_draggable: Boolean that indicates whether the annotation is draggable.
        is_resizable: Boolean that indicates whether the annotation is resizable.
        properties_: Additional properties or configuration for the annotation.
        existing_version: Fail the ingestion request if the node's version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the node
            (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert,
            so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion
            request, then the item will be skipped instead of failing the ingestion request.
        type: Direct relation pointing to the type node.
    """

    def __init__(
        self,
        external_id: str,
        *,
        id_: str,
        annotation_type: str,
        container_id: str | None = None,
        is_selectable: bool | None = None,
        is_draggable: bool | None = None,
        is_resizable: bool | None = None,
        properties_: dict | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        TypedNodeApply.__init__(self, CANVAS_INSTANCE_SPACE, external_id, existing_version, type)
        self.id_ = id_
        self.annotation_type = annotation_type
        self.container_id = container_id
        self.is_selectable = is_selectable
        self.is_draggable = is_draggable
        self.is_resizable = is_resizable
        self.properties_ = properties_


class CanvasAnnotation(_CanvasAnnotationProperties, TypedNode):
    """This represents the reading format of canvas annotation.

    It is used to when data is read from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the canvas annotation.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        id_: The unique identifier of the canvas annotation.
        annotation_type: The type of the annotation. Must be one of rectangle, ellipse, polyline, text or sticky.
        container_id: The optional ID of the container that the annotation is contained in.
        is_selectable: Boolean that indicates whether the annotation is selectable.
        is_draggable: Boolean that indicates whether the annotation is draggable.
        is_resizable: Boolean that indicates whether the annotation is resizable.
        properties_: Additional properties or configuration for the annotation.
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        id_: str,
        annotation_type: str,
        container_id: str | None = None,
        is_selectable: bool | None = None,
        is_draggable: bool | None = None,
        is_resizable: bool | None = None,
        properties_: dict | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.id_ = id_
        self.annotation_type = annotation_type
        self.container_id = container_id
        self.is_selectable = is_selectable
        self.is_draggable = is_draggable
        self.is_resizable = is_resizable
        self.properties_ = properties_

    def as_write(self) -> CanvasAnnotationApply:
        return CanvasAnnotationApply(
            self.external_id,
            id_=self.id_,
            annotation_type=self.annotation_type,
            container_id=self.container_id,
            is_selectable=self.is_selectable,
            is_draggable=self.is_draggable,
            is_resizable=self.is_resizable,
            properties_=self.properties_,
            existing_version=self.version,
            type=self.type,
        )


class _CogniteSolutionTagProperties:
    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_apps_shared", "CogniteSolutionTag", "v1")


class CogniteSolutionTagApply(_CogniteSolutionTagProperties, ExtendedTypedNodeApply):
    """This represents the writing format of Cognite solution tag.

    It is used to when data is written to CDF.

    Args:
        external_id: The external id of the Cognite solution tag.
        name: Name of the solution tag/label
        description: Description of the solution tag/label
        color: Color of the solution tag/label
        existing_version: Fail the ingestion request if the node's version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the node
            (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert,
            so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion
            request, then the item will be skipped instead of failing the ingestion request.
        type: Direct relation pointing to the type node.
    """

    def __init__(
        self,
        external_id: str,
        *,
        name: str,
        description: str | None = None,
        color: str | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        TypedNodeApply.__init__(self, SOLUTION_TAG_SPACE, external_id, existing_version, type)
        self.name = name
        self.description = description
        self.color = color


class CogniteSolutionTag(_CogniteSolutionTagProperties, TypedNode):
    """This represents the reading format of Cognite solution tag.

    It is used to when data is read from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the Cognite solution tag.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        name: Name of the solution tag/label
        description: Description of the solution tag/label
        color: Color of the solution tag/label
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        name: str,
        description: str | None = None,
        color: str | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.name = name
        self.description = description
        self.color = color

    def as_write(self) -> CogniteSolutionTagApply:
        return CogniteSolutionTagApply(
            self.external_id,
            name=self.name,
            description=self.description,
            color=self.color,
            existing_version=self.version,
            type=self.type,
        )


class _ContainerReferenceProperties:
    container_reference_type = PropertyOptions("containerReferenceType")
    resource_id = PropertyOptions("resourceId")
    id_ = PropertyOptions("id")
    resource_sub_id = PropertyOptions("resourceSubId")
    charts_id = PropertyOptions("chartsId")
    properties_ = PropertyOptions("properties")
    max_width = PropertyOptions("maxWidth")
    max_height = PropertyOptions("maxHeight")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_industrial_canvas", "ContainerReference", "v2")


class ContainerReferenceApply(_ContainerReferenceProperties, ExtendedTypedNodeApply):
    """This represents the writing format of container reference.

    It is used to when data is written to CDF.

    Args:
        external_id: The external id of the container reference.
        container_reference_type: The type of the container. Must be one of file, timeseries, asset, event, threeD
        resource_id: The ID of the CDF resource associated with the container.
        id_: The unique identifier of the container reference.
        resource_sub_id: An optional sub ID of the CDF resource. Could, for example, be the revision id of a 3D model
        charts_id: An optional id for a charts' uuid. It is needed since all the other resources are using a numeric id
            - aka 'resourceId'.
        label: The label or name associated with the container.
        properties_: Additional properties or configuration for the container.
        x: The X-coordinate of the container in pixels.
        y: The Y-coordinate of the container in pixels.
        width: The width of the container in pixels.
        height: The height of the container in pixels.
        max_width: The maximum allowed width of the container in pixels.
        max_height: The maximum allowed height of the container in pixels.
        existing_version: Fail the ingestion request if the node's version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the node
            (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert,
            so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion
            request, then the item will be skipped instead of failing the ingestion request.
        type: Direct relation pointing to the type node.
    """

    def __init__(
        self,
        external_id: str,
        *,
        container_reference_type: str,
        resource_id: int,
        id_: str | None = None,
        resource_sub_id: int | None = None,
        charts_id: str | None = None,
        label: str | None = None,
        properties_: dict | None = None,
        x: float | None = None,
        y: float | None = None,
        width: float | None = None,
        height: float | None = None,
        max_width: float | None = None,
        max_height: float | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        TypedNodeApply.__init__(self, CANVAS_INSTANCE_SPACE, external_id, existing_version, type)
        self.container_reference_type = container_reference_type
        self.resource_id = resource_id
        self.id_ = id_
        self.resource_sub_id = resource_sub_id
        self.charts_id = charts_id
        self.label = label
        self.properties_ = properties_
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.max_width = max_width
        self.max_height = max_height

    def as_asset_centric_id(self) -> AssetCentricId:
        """Returns the asset-centric ID of the container reference."""
        if self.container_reference_type not in {"file", "timeseries", "asset", "event"}:
            raise ValueError(
                f"Container reference type '{self.container_reference_type}' is not supported for asset-centric ID."
            )
        return AssetCentricId(
            self.container_reference_type,
            self.resource_id,
        )


class ContainerReference(_ContainerReferenceProperties, TypedNode):
    """This represents the reading format of container reference.

    It is used to when data is read from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the container reference.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        container_reference_type: The type of the container. Must be one of file, timeseries, asset, event, threeD
        resource_id: The ID of the CDF resource associated with the container.
        id_: The unique identifier of the container reference.
        resource_sub_id: An optional sub ID of the CDF resource. Could, for example, be the revision id of a 3D model
        charts_id: An optional id for a charts' uuid. It is needed since all the other resources are using a numeric id
            - aka 'resourceId'.
        label: The label or name associated with the container.
        properties_: Additional properties or configuration for the container.
        x: The X-coordinate of the container in pixels.
        y: The Y-coordinate of the container in pixels.
        width: The width of the container in pixels.
        height: The height of the container in pixels.
        max_width: The maximum allowed width of the container in pixels.
        max_height: The maximum allowed height of the container in pixels.
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        container_reference_type: str,
        resource_id: int,
        id_: str | None = None,
        resource_sub_id: int | None = None,
        charts_id: str | None = None,
        label: str | None = None,
        properties_: dict | None = None,
        x: float | None = None,
        y: float | None = None,
        width: float | None = None,
        height: float | None = None,
        max_width: float | None = None,
        max_height: float | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.container_reference_type = container_reference_type
        self.resource_id = resource_id
        self.id_ = id_
        self.resource_sub_id = resource_sub_id
        self.charts_id = charts_id
        self.label = label
        self.properties_ = properties_
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.max_width = max_width
        self.max_height = max_height

    def as_write(self) -> ContainerReferenceApply:
        return ContainerReferenceApply(
            self.external_id,
            container_reference_type=self.container_reference_type,
            resource_id=self.resource_id,
            id_=self.id_,
            resource_sub_id=self.resource_sub_id,
            charts_id=self.charts_id,
            label=self.label,
            properties_=self.properties_,
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            max_width=self.max_width,
            max_height=self.max_height,
            existing_version=self.version,
            type=self.type,
        )

    def as_asset_centric_id(self) -> AssetCentricId:
        """Returns the asset-centric ID of the container reference."""
        if self.container_reference_type not in {"file", "timeseries", "asset", "event"}:
            raise ValueError(
                f"Container reference type '{self.container_reference_type}' is not supported for asset-centric ID."
            )
        return AssetCentricId(
            self.container_reference_type,
            self.resource_id,
        )


class _FdmInstanceContainerReferenceProperties:
    container_reference_type = PropertyOptions("containerReferenceType")
    instance_external_id = PropertyOptions("instanceExternalId")
    instance_space = PropertyOptions("instanceSpace")
    view_external_id = PropertyOptions("viewExternalId")
    view_space = PropertyOptions("viewSpace")
    id_ = PropertyOptions("id")
    view_version = PropertyOptions("viewVersion")
    properties_ = PropertyOptions("properties")
    max_width = PropertyOptions("maxWidth")
    max_height = PropertyOptions("maxHeight")

    @classmethod
    def get_source(cls) -> ViewId:
        return ViewId("cdf_industrial_canvas", "FdmInstanceContainerReference", "v1")


class FdmInstanceContainerReferenceApply(_FdmInstanceContainerReferenceProperties, ExtendedTypedNodeApply):
    """This represents the writing format of fdm instance container reference.

    It is used to when data is written to CDF.

    Args:
        external_id: The external id of the fdm instance container reference.
        container_reference_type: The container type of the FDM instance container. Must be 'fdmInstance'
        instance_external_id: The external ID of the instance
        instance_space: The space the instance belongs to
        view_external_id: The externalId of the view used to fetch the instance
        view_space: The space of the view used to fetch the instance
        id_: The application side ID of the FDM instance container reference.
        view_version: The version of the view used to fetch the instance. If not specified, the latest version is
            assumed to be used.
        label: The label or name associated with the container.
        properties_: Additional properties or configuration for the container.
        x: The X-coordinate of the container in pixels.
        y: The Y-coordinate of the container in pixels.
        width: The width of the container in pixels.
        height: The height of the container in pixels.
        max_width: The maximum allowed width of the container in pixels.
        max_height: The maximum allowed height of the container in pixels.
        existing_version: Fail the ingestion request if the node's version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the node
            (for the specified container or node). If existingVersion is set to 0, the upsert will behave as an insert,
            so it will fail the bulk if the item already exists. If skipOnVersionConflict is set on the ingestion
            request, then the item will be skipped instead of failing the ingestion request.
        type: Direct relation pointing to the type node.
    """

    def __init__(
        self,
        external_id: str,
        *,
        container_reference_type: str,
        instance_external_id: str,
        instance_space: str,
        view_external_id: str,
        view_space: str,
        id_: str | None = None,
        view_version: str | None = None,
        label: str | None = None,
        properties_: dict | None = None,
        x: float | None = None,
        y: float | None = None,
        width: float | None = None,
        height: float | None = None,
        max_width: float | None = None,
        max_height: float | None = None,
        existing_version: int | None = None,
        type: DirectRelationReference | tuple[str, str] | None = None,
    ) -> None:
        TypedNodeApply.__init__(self, CANVAS_INSTANCE_SPACE, external_id, existing_version, type)
        self.container_reference_type = container_reference_type
        self.instance_external_id = instance_external_id
        self.instance_space = instance_space
        self.view_external_id = view_external_id
        self.view_space = view_space
        self.id_ = id_
        self.view_version = view_version
        self.label = label
        self.properties_ = properties_
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.max_width = max_width
        self.max_height = max_height


class FdmInstanceContainerReference(_FdmInstanceContainerReferenceProperties, TypedNode):
    """This represents the reading format of fdm instance container reference.

    It is used to when data is read from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the fdm instance container reference.
        version (int): DMS version.
        last_updated_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        created_time (int): The number of milliseconds since 00:00:00 Thursday, 1 January 1970,
            Coordinated Universal Time (UTC), minus leap seconds.
        container_reference_type: The container type of the FDM instance container. Must be 'fdmInstance'
        instance_external_id: The external ID of the instance
        instance_space: The space the instance belongs to
        view_external_id: The externalId of the view used to fetch the instance
        view_space: The space of the view used to fetch the instance
        id_: The application side ID of the FDM instance container reference.
        view_version: The version of the view used to fetch the instance. If not specified, the latest version is
            assumed to be used.
        label: The label or name associated with the container.
        properties_: Additional properties or configuration for the container.
        x: The X-coordinate of the container in pixels.
        y: The Y-coordinate of the container in pixels.
        width: The width of the container in pixels.
        height: The height of the container in pixels.
        max_width: The maximum allowed width of the container in pixels.
        max_height: The maximum allowed height of the container in pixels.
        type: Direct relation pointing to the type node.
        deleted_time: The number of milliseconds since 00:00:00 Thursday, 1 January 1970, Coordinated Universal Time
            (UTC), minus leap seconds. Timestamp when the instance was soft deleted. Note that deleted instances
            are filtered out of query results, but present in sync results
    """

    def __init__(
        self,
        space: str,
        external_id: str,
        version: int,
        last_updated_time: int,
        created_time: int,
        *,
        container_reference_type: str,
        instance_external_id: str,
        instance_space: str,
        view_external_id: str,
        view_space: str,
        id_: str | None = None,
        view_version: str | None = None,
        label: str | None = None,
        properties_: dict | None = None,
        x: float | None = None,
        y: float | None = None,
        width: float | None = None,
        height: float | None = None,
        max_width: float | None = None,
        max_height: float | None = None,
        type: DirectRelationReference | None = None,
        deleted_time: int | None = None,
    ) -> None:
        TypedNode.__init__(self, space, external_id, version, last_updated_time, created_time, deleted_time, type)
        self.container_reference_type = container_reference_type
        self.instance_external_id = instance_external_id
        self.instance_space = instance_space
        self.view_external_id = view_external_id
        self.view_space = view_space
        self.id_ = id_
        self.view_version = view_version
        self.label = label
        self.properties_ = properties_
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.max_width = max_width
        self.max_height = max_height

    def as_write(self) -> FdmInstanceContainerReferenceApply:
        return FdmInstanceContainerReferenceApply(
            self.external_id,
            container_reference_type=self.container_reference_type,
            instance_external_id=self.instance_external_id,
            instance_space=self.instance_space,
            view_external_id=self.view_external_id,
            view_space=self.view_space,
            id_=self.id_,
            view_version=self.view_version,
            label=self.label,
            properties_=self.properties_,
            x=self.x,
            y=self.y,
            width=self.width,
            height=self.height,
            max_width=self.max_width,
            max_height=self.max_height,
            existing_version=self.version,
            type=self.type,
        )


class IndustrialCanvasApply(CogniteResource):
    """This class represents the writing format of IndustrialCanvas.
    It is used to when data is written to CDF.
    Args:
        canvas: The Canvas object.
        annotations: A list of CanvasAnnotation objects.
        container_references: A list of ContainerReference objects.
        fdm_instance_container_references: A list of FdmInstanceContainerReference objects.
        solution_tags: A list of CogniteSolutionTag objects.
    """

    def __init__(
        self,
        canvas: CanvasApply,
        annotations: list[CanvasAnnotationApply] | None = None,
        container_references: list[ContainerReferenceApply] | None = None,
        fdm_instance_container_references: list[FdmInstanceContainerReferenceApply] | None = None,
        solution_tags: list[CogniteSolutionTagApply] | None = None,
    ) -> None:
        self.canvas = canvas
        self.annotations: list[CanvasAnnotationApply] = annotations or []
        self.container_references: list[ContainerReferenceApply] = container_references or []
        self.fdm_instance_container_references: list[FdmInstanceContainerReferenceApply] = (
            fdm_instance_container_references or []
        )
        self.solution_tags: list[CogniteSolutionTagApply] = solution_tags or []

    def as_instances(self) -> list[InstanceApply]:
        """Convert the IndustrialCanvasApply to a list of InstanceApply objects."""
        instances: list[InstanceApply] = [self.canvas]
        instances.extend(self.annotations)
        instances.extend(self.container_references)
        instances.extend(self.fdm_instance_container_references)
        instances.extend(self.solution_tags)
        for items, edge_type in [
            (self.annotations, ANNOTATION_EDGE_TYPE),
            (self.container_references, CONTAINER_REFERENCE_EDGE_TYPE),
            (self.fdm_instance_container_references, FDM_CONTAINER_REFERENCE_EDGE_TYPE),
        ]:
            # MyPy does not recognize that items is a list of TypedNodeApply
            for item in items:  # type: ignore[attr-defined]
                instances.append(
                    EdgeApply(
                        space=CANVAS_INSTANCE_SPACE,
                        external_id=f"{self.canvas.external_id}_{item.external_id}",
                        start_node=DirectRelationReference(
                            space=self.canvas.space, external_id=self.canvas.external_id
                        ),
                        end_node=DirectRelationReference(space=item.space, external_id=item.external_id),
                        type=edge_type,
                    )
                )

        return instances

    def as_id(self) -> str:
        return self.canvas.external_id

    def as_instance_ids(self, include_solution_tags: bool = False) -> list[NodeId | EdgeId]:
        """Return a list of IDs for the instances in the IndustrialCanvasApply."""
        instances = self.as_instances()
        ids: list[NodeId | EdgeId] = []
        for instance in instances:
            if isinstance(instance, NodeApply):
                if include_solution_tags or not isinstance(instance, CogniteSolutionTagApply):
                    ids.append(NodeId(instance.space, instance.external_id))
            elif isinstance(instance, EdgeApply):
                ids.append(EdgeId(instance.space, instance.external_id))
            else:
                raise TypeError(f"Unexpected instance type: {type(instance)}")
        return ids

    def dump(self, keep_existing_version: bool = True) -> dict[str, object]:
        """Dump the IndustrialCanvasApply to a dictionary."""
        return {
            "canvas": self.canvas.dump(keep_existing_version=keep_existing_version),
            "annotations": [
                annotation.dump(keep_existing_version=keep_existing_version) for annotation in self.annotations
            ],
            "containerReferences": [
                container_ref.dump(keep_existing_version=keep_existing_version)
                for container_ref in self.container_references
            ],
            "fdmInstanceContainerReferences": [
                fdm_instance_container_ref.dump(keep_existing_version=keep_existing_version)
                for fdm_instance_container_ref in self.fdm_instance_container_references
            ],
            "solutionTags": [
                solution_tag.dump(keep_existing_version=keep_existing_version) for solution_tag in self.solution_tags
            ],
        }

    def create_backup(self) -> "IndustrialCanvasApply":
        """Create a duplicate of the IndustrialCanvasApply instance."""
        new_canvas_id = str(uuid4())
        new_canvas = CanvasApply._load(self.canvas.dump(keep_existing_version=False))
        new_canvas.external_id = new_canvas_id
        new_canvas.source_canvas_id = self.canvas.external_id
        new_canvas.updated_at = datetime.now(tz=timezone.utc)
        # Solution tags are not duplicated, they are reused
        new_container = IndustrialCanvasApply(new_canvas, [], [], [], solution_tags=self.solution_tags)
        items: list[ContainerReferenceApply] | list[CanvasAnnotationApply] | list[FdmInstanceContainerReferenceApply]
        item_cls: type[CanvasAnnotationApply] | type[ContainerReferenceApply] | type[FdmInstanceContainerReferenceApply]
        new_item_list: list[NodeApply]
        for items, item_cls, new_item_list in [  # type: ignore[assignment]
            (self.annotations, CanvasAnnotationApply, new_container.annotations),
            (self.container_references, ContainerReferenceApply, new_container.container_references),
            (
                self.fdm_instance_container_references,
                FdmInstanceContainerReferenceApply,
                new_container.fdm_instance_container_references,
            ),
        ]:
            for item in items:
                # Serialize the item to create a new instance
                new_item = item_cls._load(item.dump(keep_existing_version=False))
                new_item.id_ = str(uuid4())
                new_item.external_id = f"{new_canvas_id}_{new_item.external_id}"
                new_item_list.append(new_item)
        return new_container

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> Self:
        """Load an IndustrialCanvasApply instance from a resource dictionary."""
        raise NotImplementedError("IndustrialCanvasApply cannot be loaded from dict.")


class IndustrialCanvas(WriteableCogniteResource[IndustrialCanvasApply]):
    """This class represents one instance of the Canvas with all connected data."""

    def __init__(
        self,
        canvas: Canvas,
        annotations: NodeListWithCursor[CanvasAnnotation] | None = None,
        container_references: NodeListWithCursor[ContainerReference] | None = None,
        fdm_instance_container_references: NodeListWithCursor[FdmInstanceContainerReference] | None = None,
        solution_tags: NodeListWithCursor[CogniteSolutionTag] | None = None,
    ) -> None:
        self.canvas = canvas
        self.annotations = annotations or NodeListWithCursor[CanvasAnnotation]([], None)
        self.container_references = container_references or NodeListWithCursor[ContainerReference]([], None)
        self.fdm_instance_container_references = fdm_instance_container_references or NodeListWithCursor[
            FdmInstanceContainerReference
        ]([], None)
        self.solution_tags = solution_tags or NodeListWithCursor[CogniteSolutionTag]([], None)

    @classmethod
    def _load(cls, resource: dict[str, Any], cognite_client: CogniteClient | None = None) -> "IndustrialCanvas":
        """Load an IndustrialCanvas instance from a QueryResult."""
        if not ("canvas" in resource and isinstance(resource["canvas"], Sequence) and len(resource["canvas"]) == 1):
            raise ValueError("Resource does not contain a canvas node.")
        canvas_resource = resource["canvas"][0]
        if isinstance(canvas_resource, dict):
            canvas = Canvas._load(canvas_resource)
        elif isinstance(canvas_resource, Canvas):
            canvas = canvas_resource
        elif isinstance(canvas_resource, Node):
            canvas = Canvas._load(canvas_resource.dump())
        else:
            raise TypeError(f"Canvas resource {type(canvas_resource)} is not supported.")
        return cls(
            canvas=canvas,
            annotations=cls._load_items(resource.get("annotations"), CanvasAnnotation),
            container_references=cls._load_items(resource.get("containerReferences"), ContainerReference),
            fdm_instance_container_references=cls._load_items(
                resource.get("fdmInstanceContainerReferences"), FdmInstanceContainerReference
            ),
            solution_tags=cls._load_items(resource.get("solutionTags"), CogniteSolutionTag),
        )

    @classmethod
    def _load_items(cls, items: object | None, node_cls: type[T_Node]) -> NodeListWithCursor[T_Node]:
        if items is None:
            return NodeListWithCursor[T_Node]([], None)
        elif isinstance(items, Sequence):
            nodes: list[T_Node] = []
            for node in items:
                if isinstance(node, dict):
                    # Bug in PySDK node_cls._load returns an instances of T_Node
                    nodes.append(node_cls._load(node))  # type: ignore[arg-type]
                elif isinstance(node, node_cls):
                    nodes.append(node)
                elif isinstance(node, Node):
                    # Bug in PySDK, node_cls._load returns an instance of T_Node
                    nodes.append(node_cls._load(node.dump()))  # type: ignore[arg-type]
                else:
                    raise TypeError(f"Expected a sequence of {node_cls.__name__}, got {type(node).__name__}")
            return NodeListWithCursor[T_Node](
                nodes,
                items.cursor if isinstance(items, NodeListWithCursor) else None,
            )
        raise TypeError(f"Expected a sequence of {node_cls.__name__}, got {type(items).__name__}")

    def dump(self, camel_case: bool = True) -> dict[str, list]:
        """Dump the IndustrialCanvas to a dictionary."""
        if not camel_case:
            raise NotImplementedError("snake_case dump is not supported")
        return {
            "canvas": [self.canvas.dump()],
            "annotations": [annotation.dump() for annotation in self.annotations],
            "containerReferences": [container_ref.dump() for container_ref in self.container_references],
            "fdmInstanceContainerReferences": [
                fdm_instance_container_ref.dump()
                for fdm_instance_container_ref in self.fdm_instance_container_references
            ],
            "solutionTags": [solution_tag.dump() for solution_tag in self.solution_tags],
        }

    def as_write(self) -> "IndustrialCanvasApply":
        return IndustrialCanvasApply(
            canvas=self.canvas.as_write(),
            annotations=[annotation.as_write() for annotation in self.annotations],
            container_references=[container_ref.as_write() for container_ref in self.container_references],
            fdm_instance_container_references=[
                fdm_instance_container_ref.as_write()
                for fdm_instance_container_ref in self.fdm_instance_container_references
            ],
            solution_tags=[solution_tag.as_write() for solution_tag in self.solution_tags],
        )

    def as_id(self) -> str:
        return self.canvas.external_id


class IndustrialCanvasApplyList(CogniteResourceList):
    _RESOURCE = IndustrialCanvasApply


class IndustrialCanvasList(WriteableCogniteResourceList[IndustrialCanvasApply, IndustrialCanvas]):
    _RESOURCE = IndustrialCanvas

    def as_write(self) -> CogniteResourceList[IndustrialCanvasApply]:
        return IndustrialCanvasApplyList([item.as_write() for item in self])



import re
from pathlib import Path

import pytest

from cognite_toolkit import _version
from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.constants import RESOURCES_PATH
from cognite_toolkit._cdf_tk.exceptions import ToolkitTOMLFormatError
from tests.constants import REPO_ROOT


# Reset singleton before each test to ensure test isolation
@pytest.fixture(autouse=True)
def reset_cdf_toml_singleton():
    global _CDF_TOML
    _CDF_TOML = None
    yield
    _CDF_TOML = None  # Clean up after test as well


class TestCDFToml:
    def test_load_repo_root_config(self) -> None:
        config = CDFToml.load(REPO_ROOT)

        assert config.modules.version == _version.__version__

    @pytest.mark.parametrize(
        "invalid_toml_content, expected_error_message",
        [
            # Invalid: package type is unknown
            (
                """
                [cdf]
                [modules]
                version = "0.0.0"
                [library.bad_lib]
                url = "ftp://bad.com"
                checksum = "1234567890abcdef"
                """,
                "Invalid library configuration for 'bad_lib': URL must start with 'https'",
            ),
            # Invalid: library missing url for https type
            (
                """
                [cdf]
                [modules]
                version = "0.0.0"
                [library.missing_url]
                checksum = "1234567890abcdef"
                """,
                "Invalid library configuration for 'missing_url': Library configuration must contain 'url' field.",
            ),
            # Invalid: library url is not a valid URL
            (
                """
                [cdf]
                [modules]
                version = "0.0.0"
                [library.invalid_url]
                url = "bad.com"
                checksum = "1234567890abcdef"
                """,
                "Invalid library configuration for 'invalid_url': URL is missing scheme or network location (e.g., 'https://domain.com')",
            ),
            # Invalid: library url does not end with .zip
            (
                """
                [cdf]
                [modules]
                version = "0.0.0"
                [library.invalid_zip]
                url = "https://example.com/my-package/my-package.txt"
                checksum = "1234567890abcdef"
                """,
                "Invalid library configuration for 'invalid_zip': URL must point to a .zip file.",
            ),
            # Invalid: library checksum is not a valid hex string
            (
                """
                [cdf]
                [modules]
                version = "0.0.0"
                [library.invalid_checksum]
                url = "https://example.com/my-package/my-package.zip"
                """,
                "Library configuration must contain 'checksum' field",
            ),
        ],
        ids=[
            "unknown_package_type",
            "missing_url",
            "invalid_url",
            "invalid_zip",
            "invalid_checksum",
        ],
    )
    def test_load_invalid_toml_content(self, tmp_path: Path, invalid_toml_content: str, expected_error_message: str):
        file_path = tmp_path / CDFToml.file_name
        file_path.write_text(invalid_toml_content)

        with pytest.raises(ToolkitTOMLFormatError, match=re.escape(expected_error_message)):
            CDFToml.load(cwd=tmp_path, use_singleton=False)

    def test_load_package_url(self, tmp_path: Path):
        valid_toml_content = """
        [cdf]
        [modules]
        version = "0.0.0"
        [library.valid_url]
        url = "https://github.com/cognitedata/package/archive/refs/tags/0.0.1.zip"
        checksum = "1234567890abcdef"
        """
        file_path = tmp_path / CDFToml.file_name
        file_path.write_text(valid_toml_content)

        config = CDFToml.load(cwd=tmp_path, use_singleton=False)
        assert config.libraries["valid_url"].url == "https://github.com/cognitedata/package/archive/refs/tags/0.0.1.zip"

    def test_default_resources_cdf_toml_has_valid_library_config(self) -> None:
        """Test that the default cdf.toml in resources has valid library configuration."""
        default_cdf_toml = CDFToml.load(cwd=RESOURCES_PATH, use_singleton=False)

        # Verify the toolkit-data library exists
        assert "toolkit-data" in default_cdf_toml.libraries
        library = default_cdf_toml.libraries["toolkit-data"]

        # Verify the URL is a valid HTTPS URL pointing to toolkit-data
        assert library.url.startswith("https://github.com/cognitedata/toolkit-data")
        assert library.url.endswith(".zip")

        # Verify the checksum has the correct format
        assert library.checksum.startswith("sha256:")



import os
import platform
import sys
import tempfile
import threading
import uuid
from collections import Counter
from contextlib import suppress
from functools import cached_property
from pathlib import Path
from typing import Any

from mixpanel import Consumer, Mixpanel, MixpanelException

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.constants import IN_BROWSER
from cognite_toolkit._cdf_tk.data_classes import CommandTrackingInfo
from cognite_toolkit._cdf_tk.tk_warnings import ToolkitWarning, WarningList
from cognite_toolkit._cdf_tk.utils import get_cicd_environment
from cognite_toolkit._version import __version__

_COGNITE_TOOLKIT_MIXPANEL_TOKEN: str = "9afc120ac61d408c81009ea7dd280a38"


class Tracker:
    def __init__(self, skip_tracking: bool = False) -> None:
        self.user_command = "".join(sys.argv[1:])
        self.mp = Mixpanel(_COGNITE_TOOLKIT_MIXPANEL_TOKEN, consumer=Consumer(api_host="api-eu.mixpanel.com"))
        self._opt_status_file = Path(tempfile.gettempdir()) / "tk-opt-status.bin"
        self.skip_tracking = self.opted_out or skip_tracking
        self._cdf_toml = CDFToml.load()

    @cached_property
    def _opt_status(self) -> str:
        if self._opt_status_file.exists():
            return self._opt_status_file.read_text()
        return ""

    @property
    def opted_out(self) -> bool:
        return self._opt_status == "opted-out"

    @property
    def opted_in(self) -> bool:
        return self._opt_status == "opted-in"

    def track_cli_command(
        self,
        warning_list: WarningList[ToolkitWarning],
        result: str | Exception,
        cmd: str,
        additional_tracking_info: CommandTrackingInfo | None = None,
    ) -> bool:
        warning_count = Counter([type(w).__name__ for w in warning_list])

        warning_details: dict[str, str | int] = {}
        for no, (warning, count) in enumerate(warning_count.most_common(3), 1):
            warning_details[f"warningMostCommon{no}Count"] = count
            warning_details[f"warningMostCommon{no}Name"] = warning

        subcommands, optional_args = self._parse_sys_args()
        event_information = {
            "userInput": self.user_command,
            "toolkitVersion": __version__,
            "$os": platform.system(),
            "pythonVersion": platform.python_version(),
            "CICD": self._cicd,
            "warningTotalCount": len(warning_list),
            **warning_details,
            "result": type(result).__name__ if isinstance(result, Exception) else result,
            "error": str(result) if isinstance(result, Exception) else "",
            "subcommands": subcommands,
            **optional_args,
            "alphaFlags": [name for name, value in self._cdf_toml.alpha_flags.items() if value],
            "plugins": [name for name, value in self._cdf_toml.plugins.items() if value],
        }

        if additional_tracking_info:
            event_information.update(additional_tracking_info.to_dict())

        return self._track(f"command{cmd.capitalize()}", event_information)

    def _track(self, event_name: str, event_information: dict[str, Any]) -> bool:
        if self.skip_tracking or not self.opted_in or "PYTEST_CURRENT_TEST" in os.environ:
            return False

        distinct_id = self.get_distinct_id()

        def track() -> None:
            # If we are unable to connect to Mixpanel, we don't want to crash the program
            with suppress(ConnectionError, MixpanelException):
                self.mp.track(
                    distinct_id,
                    event_name,
                    event_information,
                )

        if IN_BROWSER:
            # Pyodide does not support threading
            track()
        else:
            thread = threading.Thread(
                target=track,
                daemon=False,
            )
            thread.start()

        return True

    def get_distinct_id(self) -> str:
        cache = Path(tempfile.gettempdir()) / "tk-distinct-id.bin"
        cicd = self._cicd
        if cache.exists():
            return cache.read_text()

        distinct_id = f"{cicd}-{platform.system()}-{platform.python_version()}-{uuid.uuid4()!s}"
        cache.write_text(distinct_id)
        with suppress(ConnectionError, MixpanelException):
            self.mp.people_set(
                distinct_id,
                {
                    "$os": platform.system(),
                    "$python_version": platform.python_version(),
                    "$distinct_id": distinct_id,
                    "CICD": self._cicd,
                },
            )
        return distinct_id

    @staticmethod
    def _parse_sys_args() -> tuple[list[str], dict[str, str | bool]]:
        optional_args: dict[str, str | bool] = {}
        subcommands: list[str] = []
        last_key: str | None = None
        if sys.argv and len(sys.argv) > 1:
            for arg in sys.argv[1:]:
                if arg.startswith("--") and "=" in arg:
                    if last_key:
                        optional_args[last_key] = True
                    key, value = arg.split("=", maxsplit=1)
                    optional_args[key.removeprefix("--")] = value
                elif arg.startswith("--"):
                    if last_key:
                        optional_args[last_key] = True
                    last_key = arg.removeprefix("--")
                elif last_key:
                    optional_args[last_key] = arg
                    last_key = None
                else:
                    subcommands.append(arg)

            if last_key:
                optional_args[last_key] = True
        return subcommands, optional_args

    @property
    def _cicd(self) -> str:
        return get_cicd_environment()

    def enable(self) -> None:
        self._opt_status_file.write_text("opted-in")

    def disable(self) -> None:
        self._opt_status_file.write_text("opted-out")



from collections.abc import Iterable
from datetime import datetime
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.instance import EdgeYAML, NodeYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources

_EDGE_TEST_DATA = {
    "space": "test_space",
    "externalId": "edge_1",
    "type": {"space": "type_space", "externalId": "type_edge"},
    "startNode": {"space": "node_space", "externalId": "start_node"},
    "endNode": {"space": "node_space", "externalId": "end_node"},
}


def invalid_edge_test_cases() -> Iterable:
    yield pytest.param(
        {
            "existingVersion": 1,
        },
        {
            "Missing required field: 'externalId'",
            "Missing required field: 'space'",
            "Missing required field: 'type'",
            "Missing required field: 'startNode'",
            "Missing required field: 'endNode'",
        },
        id="missing-required-fields",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "existingVersion": -1,
        },
        {"In field existingVersion input should be greater than or equal to 0"},
        id="negative-existing-version",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "unknownField": "value",
        },
        {"Unused field: 'unknownField'"},
        id="unknown-field-present",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "space": "1invalid_space",
        },
        {"In field space string should match pattern '^[a-zA-Z][a-zA-Z0-9_-]{0,41}[a-zA-Z0-9]?$'"},
        id="invalid-space-pattern",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "space": "a" * 44,
        },
        {"In field space string should have at most 43 characters"},
        id="space-too-long",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "externalId": "",
        },
        {"In field externalId string should have at least 1 character"},
        id="invalid-external-id",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "externalId": "a" * 257,
        },
        {"In field externalId string should have at most 256 characters"},
        id="external-id-too-long",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "type": {
                "space": "type_space",
                "externalId": "type_edge",
                "unknownField": "value",
            },
        },
        {"In type unused field: 'unknownField'"},
        id="unknown-field-node-type",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "type": {
                "space": "",
                "externalId": "type_edge",
            },
        },
        {"In type.space string should have at least 1 character"},
        id="invalid-space-node-type",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "sources": [
                {
                    "source": {"type": "view", "space": "space1", "externalId": "view1", "version": "1"},
                    "properties": {23: "value1"},
                },
            ],
        },
        {
            "In sources[1].properties[24][key] input should be a valid string. Got 23 of type int. Hint: Use double quotes to force string."
        },
        id="invalid-properties-key",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "sources": [
                {
                    "source": {"type": "view", "space": "space1", "externalId": "view1", "version": "1"},
                    "properties": {"prop1": datetime.now()},
                },
            ],
        },
        {"In sources[1].properties.prop1 input was not a valid JSON value"},
        id="invalid-properties-value",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "sources": [
                {
                    "source": {"type": "unknown", "space": "space1", "externalId": "view1"},
                    "properties": {"prop1": "value"},
                },
            ],
        },
        {
            "In sources[1].source.ViewReference.type input should be 'view'. Got 'unknown'.",
            "In sources[1].source.ViewReference missing required field: 'version'",
            "In sources[1].source.ContainerReference.type input should be 'container'. Got 'unknown'.",
        },
        id="invalid-source-type",
    )


def valid_edge_test_cases() -> Iterable:
    yield pytest.param(
        _EDGE_TEST_DATA,
        id="minimal-valid-edge",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "sources": [
                {
                    "source": {"type": "view", "space": "view_space", "externalId": "view_1", "version": "1"},
                    "properties": {"name": "Test Edge", "weight": 1.23, "tags": ["tag1", "tag2"]},
                }
            ],
            "existingVersion": 1,
        },
        id="full-edge-with-all-fields",
    )
    yield pytest.param(
        {
            **_EDGE_TEST_DATA,
            "sources": [
                {
                    "source": {"type": "view", "space": "space1", "externalId": "view1", "version": "1"},
                    "properties": {"prop1": "value1"},
                },
                {
                    "source": {"type": "view", "space": "space2", "externalId": "view2", "version": "1"},
                    "properties": {"prop2": "value2"},
                },
            ],
        },
        id="multi-source-edge",
    )


class TestNodeYAML:
    """Test suite for NodeYAML class."""

    @pytest.mark.parametrize("data", list(find_resources("Node")))
    def test_load_valid_node_from_resources(self, data: dict[str, object]) -> None:
        """Test loading valid nodes from resource files."""
        loaded = NodeYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    def test_node_with_minimal_fields(self) -> None:
        """Test node with only required fields."""
        minimal_data = {
            "space": "test_space",
            "externalId": "test_node",
        }
        loaded = NodeYAML.model_validate(minimal_data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == minimal_data

    def test_node_with_all_fields(self) -> None:
        """Test node with all fields populated."""
        full_data = {
            "space": "test_space",
            "externalId": "test_node",
            "type": {"space": "type_space", "externalId": "type_node"},
            "sources": [
                {
                    "source": {"type": "view", "space": "view_space", "externalId": "view_1", "version": "1"},
                    "properties": {"name": "Test Node", "value": 42, "tags": ["tag1", "tag2"]},
                }
            ],
            "existingVersion": 1,
        }
        loaded = NodeYAML.model_validate(full_data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == full_data

    def test_invalid_node_missing_required_fields(self) -> None:
        """Test that missing required fields are properly reported."""
        invalid_data = {
            "space": "my_space"
            # Missing 'externalId'
        }

        warning_list = validate_resource_yaml_pydantic(invalid_data, NodeYAML, Path("test.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == {"Missing required field: 'externalId'"}

    def test_node_with_multiple_sources(self) -> None:
        """Test node with multiple view sources."""
        data = {
            "space": "my_space",
            "externalId": "multi_source_node",
            "sources": [
                {
                    "source": {"type": "view", "space": "space1", "externalId": "view1", "version": "1"},
                    "properties": {"prop1": "value1"},
                },
                {
                    "source": {"type": "view", "space": "space2", "externalId": "view2", "version": "1"},
                    "properties": {"prop2": "value2"},
                },
            ],
        }
        loaded = NodeYAML.model_validate(data)
        assert len(loaded.sources) == 2
        assert isinstance(loaded.sources[0].properties, dict)
        assert "prop1" in loaded.sources[0].properties
        assert loaded.sources[0].properties["prop1"] == "value1"
        assert isinstance(loaded.sources[1].properties, dict)
        assert "prop2" in loaded.sources[1].properties
        assert loaded.sources[1].properties["prop2"] == "value2"


class TestEdgeYAML:
    """Test suite for EdgeYAML class."""

    @pytest.mark.parametrize("data", list(find_resources("Edge")))
    def test_load_valid_edge_from_resources(self, data: dict[str, object]) -> None:
        """Test loading valid edges from resource files."""
        loaded = EdgeYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", list(valid_edge_test_cases()))
    def test_valid_edges(self, data: dict[str, object]) -> None:
        """Test valid edge configurations."""
        loaded = EdgeYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_edge_test_cases()))
    def test_invalid_edge_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        """Test invalid edges with error messages."""
        warning_list = validate_resource_yaml_pydantic(data, EdgeYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert len(format_warning.errors) == len(expected_errors)
        assert set(format_warning.errors) == expected_errors



import time

from cognite.client.data_classes import FileMetadata, FileMetadataWrite
from cognite.client.data_classes.data_modeling import NodeApplyResultList, Space
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteFileApply

from cognite_toolkit._cdf_tk.client import ToolkitClient
from tests.test_integration.constants import RUN_UNIQUE_ID


class TestExtendedFilesAPI:
    def test_set_pending_instance_id(self, dev_cluster_client: ToolkitClient, dev_space: str) -> None:
        """Happy path for setting a pending instance ID on a file

        1. Create file with content.
        3. Set pending instance ID.
        4. Create a CogniteFile
        5. Retrieve file content using the Node ID.
        """
        client = dev_cluster_client
        metadata = FileMetadataWrite(
            external_id=f"ts_toolkit_integration_test_happy_path_files_{RUN_UNIQUE_ID}",
            name="Toolkit Integration Test Happy Path Files",
            mime_type="text/plain",
        )
        cognite_file = CogniteFileApply(
            space=dev_space,
            external_id=metadata.external_id,
            name="Toolkit Integration Test Happy Path",
            mime_type="text/plain",
        )
        content = b"Hello, this is a test file's content."
        created: FileMetadata | None = None
        created_dm: NodeApplyResultList | None = None
        try:
            created, _ = client.files.create(metadata)
            client.files.upload_content_bytes(content, external_id=created.external_id)

            updated = client.files.set_pending_ids(cognite_file.as_id(), id=created.id)
            assert updated.pending_instance_id == cognite_file.as_id()

            created_dm = client.data_modeling.instances.apply(cognite_file).nodes

            downloaded_bytes = client.files.download_bytes(instance_id=cognite_file.as_id())

            assert downloaded_bytes == content
        finally:
            if created is not None and created_dm is None:
                client.files.delete(external_id=metadata.external_id)
            if created_dm is not None:
                # This will delete the CogniteFile and the asset-centric file
                client.data_modeling.instances.delete(cognite_file.as_id())

    def test_unlink_instance_ids(self, toolkit_client_with_pending_ids: ToolkitClient, toolkit_space: Space) -> None:
        client = toolkit_client_with_pending_ids
        space = toolkit_space.space
        metadata = FileMetadataWrite(
            external_id=f"file_toolkit_integration_test_unlink_{RUN_UNIQUE_ID}",
            name="Toolkit Integration Test Unlink",
            mime_type="text/plain",
        )
        cognite_file = CogniteFileApply(
            space=space,
            external_id=metadata.external_id,
            name="Toolkit Integration Test Unlink",
        )
        content = b"Hello, this is a test file's content."
        created: FileMetadata | None = None
        created_dm: NodeApplyResultList | None = None
        try:
            created, _ = client.files.create(metadata)
            client.files.upload_content_bytes(content, external_id=created.external_id)

            updated = client.files.set_pending_ids(cognite_file.as_id(), id=created.id)

            assert updated.pending_instance_id == cognite_file.as_id()

            created_dm = client.data_modeling.instances.apply(cognite_file).nodes

            retrieved_ts: FileMetadata | None = None
            for _ in range(30):  # Wait up to 30 seconds for the syncer to update the file metadata
                retrieved_ts = client.files.retrieve(instance_id=cognite_file.as_id())
                if retrieved_ts is not None:
                    break
                time.sleep(1)  # Wait for the syncer to update the file metadata

            assert retrieved_ts is not None, "File was not linked to instance within timeout"
            assert retrieved_ts.id == created.id

            unlinked = client.files.unlink_instance_ids(id=created.id)
            assert unlinked.id == created.id

            client.data_modeling.instances.delete(cognite_file.as_id())
            created_dm = None

            # Still existing asset-centric file.
            retrieved_ts = client.files.retrieve(external_id=metadata.external_id)
            assert retrieved_ts is not None
            assert retrieved_ts.id == created.id
        finally:
            if created is not None and created_dm is None:
                client.files.delete(external_id=metadata.external_id, ignore_unknown_ids=True)
            if created_dm is not None:
                client.data_modeling.instances.delete(cognite_file.as_id())



from dataclasses import dataclass
from enum import Enum

from cognite_toolkit._cdf_tk.cdf_toml import CDFToml
from cognite_toolkit._cdf_tk.constants import clean_name


@dataclass
class Plugin:
    name: str
    description: str

    def is_enabled(self) -> bool:
        return CDFToml.load().plugins.get(clean_name(self.name), False)


class Plugins(Enum):
    run = Plugin("run", "plugin for Run command to execute Python scripts in CDF")
    dump = Plugin("dump", "plugin for Dump command to retrieve Asset resources from CDF")
    purge = Plugin("purge", "plugin for Purge command to remove datasets and spaces from CDF")
    dev = Plugin("dev", "plugin for commands to develop modules in CDF")

    @staticmethod
    def list() -> dict[str, bool]:
        res = {plugin.name: Plugin.is_enabled(plugin.value) for plugin in Plugins}
        return res



from cognite.client import ClientConfig, CogniteClient
from cognite.client._api_client import APIClient

from .capabilities import CapabilitiesAPI
from .data_postprocessing import DataPostProcessingAPI
from .frames import FramesAPI
from .locations import LocationsAPI
from .maps import MapsAPI
from .robots import RobotsAPI


class RoboticsAPI(APIClient):
    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self.robots = RobotsAPI(config, api_version, cognite_client)
        self.data_postprocessing = DataPostProcessingAPI(config, api_version, cognite_client)
        self.locations = LocationsAPI(config, api_version, cognite_client)
        self.frames = FramesAPI(config, api_version, cognite_client)
        self.maps = MapsAPI(config, api_version, cognite_client)
        self.capabilities = CapabilitiesAPI(config, api_version, cognite_client)



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.sequence import SequenceRowYAML, SequenceYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def invalid_sequence_test_cases() -> Iterable:
    yield pytest.param(
        {"name": "Sequence 1"},
        {"Missing required field: 'externalId'", "Missing required field: 'columns'"},
        id="missing-required-fields",
    )
    yield pytest.param(
        {"name": "Sequence 1", "columns": []},
        {
            "In field columns list should have at least 1 item after validation, not 0",
            "Missing required field: 'externalId'",
        },
        id="columns-list-validation-errors",
    )
    yield pytest.param(
        {
            "externalId": "seq_1",
            "columns": [
                {"name": "Column 1"},
            ],
        },
        {"In columns[1] missing required field: 'externalId'"},
        id="missing-externalId-in-column",
    )
    yield pytest.param(
        {
            "externalId": "seq_1",
            "columns": [
                {"externalId": "col_1", "valueType": "INVALID_TYPE"},
            ],
        },
        {
            "In columns[1].valueType input should be 'STRING', 'string', 'DOUBLE', 'double', 'LONG' or 'long'. Got 'INVALID_TYPE'."
        },
        id="invalid-valueType-in-column",
    )
    yield pytest.param(
        {"externalId": "seq_1", "columns": [{"externalId": "col_1"}], "unknownField": "value"},
        {"Unused field: 'unknownField'"},
        id="unknown-field-present",
    )


def valid_sequence_test_cases() -> Iterable:
    yield pytest.param(
        {
            "externalId": "minimal_sequence",
            "columns": [{"externalId": "col_1"}],
        },
        id="minimal-valid-sequence",
    )
    yield pytest.param(
        {
            "externalId": "full_sequence",
            "name": "Full Sequence",
            "description": "A complete sequence with all fields",
            "dataSetExternalId": "my_dataset",
            "assetExternalId": "my_asset",
            "metadata": {"key1": "value1", "key2": "value2"},
            "columns": [
                {
                    "externalId": "string_col",
                    "name": "String Column",
                    "description": "A string column",
                    "valueType": "STRING",
                    "metadata": {"type": "text"},
                },
                {
                    "externalId": "double_col",
                    "name": "Double Column",
                    "description": "A double precision column",
                    "valueType": "DOUBLE",
                },
                {
                    "externalId": "long_col",
                    "name": "Long Column",
                    "description": "A long integer column",
                    "valueType": "LONG",
                },
                {
                    "externalId": "string_col_2",
                    "name": "String Column",
                    "description": "A string column",
                    "valueType": "string",
                },
                {
                    "externalId": "long_col_2",
                    "name": "Long Column",
                    "description": "A long integer column",
                    "valueType": "long",
                },
                {
                    "externalId": "double_col_2",
                    "name": "Double Column",
                    "description": "A double precision column",
                    "valueType": "double",
                },
            ],
        },
        id="full-sequence-with-all-fields",
    )


class TestSequenceYAML:
    @pytest.mark.parametrize("data", list(find_resources("Sequence")))
    def test_load_valid_sequence_from_resources(self, data: dict[str, object]) -> None:
        """Test loading valid sequences from resource files."""
        loaded = SequenceYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", list(valid_sequence_test_cases()))
    def test_valid_sequences(self, data: dict[str, object]) -> None:
        """Test various valid sequence configurations."""
        loaded = SequenceYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_sequence_test_cases()))
    def test_invalid_sequence_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        """Test that invalid sequences are properly rejected with appropriate error messages."""
        warning_list = validate_resource_yaml_pydantic(data, SequenceYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == expected_errors


def invalid_sequence_row_test_cases() -> Iterable:
    yield pytest.param(
        {"externalId": "seq_row_1"},
        {"Missing required field: 'columns'", "Missing required field: 'rows'"},
        id="missing-required-fields",
    )
    yield pytest.param(
        {"externalId": "seq_row_1", "columns": [], "rows": []},
        {
            "In field columns list should have at least 1 item after validation, not 0",
            "In field rows list should have at least 1 item after validation, not 0",
        },
        id="empty-lists-validation-errors",
    )
    yield pytest.param(
        {
            "externalId": "seq_row_1",
            "columns": ["col1"],
            "rows": [{"values": [1, 2, 3]}],
        },
        {"In rows[1] missing required field: 'rowNumber'"},
        id="missing-rowNumber-in-row",
    )
    yield pytest.param(
        {
            "externalId": "seq_row_1",
            "columns": ["col1"],
            "rows": [{"rowNumber": -1, "values": [1, 2, 3]}],
        },
        {"In rows[1].rowNumber input should be greater than or equal to 0"},
        id="negative-rowNumber",
    )
    yield pytest.param(
        {
            "externalId": "seq_row_1",
            "columns": ["col1"],
            "rows": [{"rowNumber": 0, "values": []}],
        },
        {"In rows[1].values list should have at least 1 item after validation, not 0"},
        id="empty-values-in-row",
    )
    yield pytest.param(
        {
            "externalId": "seq_row_1",
            "columns": ["col1", "col2"],
            "rows": [{"rowNumber": 0, "values": [1]}],
        },
        {
            "Row number 0 has 1 value(s). Each row must have exactly 2 value(s) which is the same as the number of column(s)."
        },
        id="values-columns-count-mismatch",
    )
    yield pytest.param(
        {
            "externalId": "seq_row_1",
            "columns": ["col1"],
            "rows": [{"rowNumber": 0, "values": [1, 2, 3]}],
        },
        {
            "Row number 0 has 3 value(s). Each row must have exactly 1 value(s) which is the same as the number of column(s)."
        },
        id="too-many-values-for-columns",
    )
    yield pytest.param(
        {
            "externalId": "seq_row_1",
            "columns": ["col1"],
            "rows": [{"rowNumber": 0, "values": [1]}],
            "unknownField": "value",
        },
        {"Unused field: 'unknownField'"},
        id="unknown-field-present",
    )
    yield pytest.param(
        {
            "externalId": "a" * 257,
            "columns": ["col1"],
            "rows": [{"rowNumber": 0, "values": [1]}],
        },
        {"In field externalId string should have at most 256 characters"},
        id="externalId-too-long",
    )
    yield pytest.param(
        {
            "externalId": "seq_row_1",
            "columns": ["a"] * 201,
            "rows": [{"rowNumber": 0, "values": [1]}],
        },
        {"In field columns list should have at most 200 items after validation, not 201"},
        id="too-many-columns",
    )
    yield pytest.param(
        {
            "externalId": "seq_row_1",
            "columns": ["col1"],
            "rows": [{"rowNumber": 0, "values": [1]}] * 10001,
        },
        {"In field rows list should have at most 10000 items after validation, not 10001"},
        id="too-many-rows",
    )


def valid_sequence_row_test_cases() -> Iterable:
    yield pytest.param(
        {
            "externalId": "minimal_sequence_row",
            "columns": ["col1"],
            "rows": [{"rowNumber": 0, "values": [1]}],
        },
        id="minimal-valid-sequence-row",
    )
    yield pytest.param(
        {
            "externalId": "full_sequence_row",
            "columns": ["string_col", "double_col", "long_col"],
            "rows": [
                {"rowNumber": 0, "values": ["text1", 1.5, 100]},
                {"rowNumber": 2, "values": [None, None, None]},
                {"rowNumber": 3, "values": ["text3", 3.14, 300]},
            ],
        },
        id="full-sequence-row-with-multiple-rows-and-types",
    )


class TestSequenceRowYAML:
    @pytest.mark.parametrize("data", list(find_resources("SequenceRow")))
    def test_load_valid_sequence_row_file(self, data: dict[str, object]) -> None:
        """Test loading valid sequence rows from resource files."""
        loaded = SequenceRowYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", list(valid_sequence_row_test_cases()))
    def test_valid_sequence_rows(self, data: dict[str, object]) -> None:
        """Test various valid sequence row configurations."""
        loaded = SequenceRowYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_sequence_row_test_cases()))
    def test_invalid_sequence_row(self, data: dict | list, expected_errors: set[str]) -> None:
        """Test that invalid sequence rows are properly rejected with appropriate error messages."""
        warning_list = validate_resource_yaml_pydantic(data, SequenceRowYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == expected_errors



from __future__ import annotations

from cognite_toolkit._cdf_tk.data_classes._module_toml import ModuleToml
from cognite_toolkit._cdf_tk.data_classes._packages import Package


class TestPackages:
    def test_package_creation_with_id(self) -> None:
        package = Package(name="test", title="Test Package", id="test-123")
        assert package.id == "test-123"

    def test_package_load_with_id(self) -> None:
        package_data = {"title": "Test Package", "id": "test-456"}
        loaded_package = Package.load("test", package_data)
        assert loaded_package.id == "test-456"

    def test_package_load_without_id(self) -> None:
        package_data = {"title": "Test Package"}
        loaded_package = Package.load("test", package_data)
        assert loaded_package.id is None

    def test_module_toml_creation_with_id(self) -> None:
        module_toml = ModuleToml(title="Test Module", id="module-123")
        assert module_toml.id == "module-123"

    def test_module_toml_load_with_id(self) -> None:
        data = {"module": {"title": "Test Module", "id": "module-456"}}
        loaded_toml = ModuleToml.load(data)
        assert loaded_toml.id == "module-456"

    def test_module_toml_load_without_id(self) -> None:
        data = {"module": {"title": "Test Module"}}
        loaded_toml = ModuleToml.load(data)
        assert loaded_toml.id is None



from collections.abc import Sequence
from typing import overload

from cognite.client import ClientConfig, CogniteClient
from cognite.client._api.files import FilesAPI
from cognite.client.data_classes.data_modeling import NodeId
from cognite.client.utils._auxiliary import split_into_chunks, unpack_items_in_payload
from cognite.client.utils._concurrency import execute_tasks
from cognite.client.utils._identifier import IdentifierSequence
from cognite.client.utils.useful_types import SequenceNotStr

from cognite_toolkit._cdf_tk.client.data_classes.extended_filemetadata import (
    ExtendedFileMetadata,
    ExtendedFileMetadataList,
)
from cognite_toolkit._cdf_tk.client.data_classes.pending_instances_ids import PendingInstanceId


class ExtendedFileMetadataAPI(FilesAPI):
    """Extended FileMetadata to include pending ID methods."""

    def __init__(self, config: ClientConfig, api_version: str | None, cognite_client: CogniteClient) -> None:
        super().__init__(config, api_version, cognite_client)
        self._PENDING_IDS_LIMIT = 1000

    @overload
    def set_pending_ids(
        self, instance_id: NodeId | tuple[str, str], id: int | None = None, external_id: str | None = None
    ) -> ExtendedFileMetadata: ...

    @overload
    def set_pending_ids(self, instance_id: Sequence[PendingInstanceId]) -> ExtendedFileMetadataList: ...

    def set_pending_ids(
        self,
        instance_id: NodeId | tuple[str, str] | Sequence[PendingInstanceId],
        id: int | None = None,
        external_id: str | None = None,
    ) -> ExtendedFileMetadata | ExtendedFileMetadataList:
        """Set a pending identifier for one or more filemetadata.

        Args:
            instance_id: The pending instance ID to set.
            id: The ID of the files.
            external_id: The external ID of the files.

        Returns:
            ExtendedFileMetadata: If a single instance ID is provided, returns the updated ExtendedFileMetadata object.
        """
        if isinstance(instance_id, NodeId) or (
            isinstance(instance_id, tuple)
            and len(instance_id) == 2
            and isinstance(instance_id[0], str)
            and isinstance(instance_id[1], str)
        ):
            return self._set_pending_ids([PendingInstanceId(NodeId.load(instance_id), id=id, external_id=external_id)])[  # type: ignore[return-value]
                0
            ]
        elif isinstance(instance_id, Sequence) and all(isinstance(item, PendingInstanceId) for item in instance_id):
            return self._set_pending_ids(instance_id)  # type: ignore[arg-type]
        else:
            raise TypeError(
                "instance_id must be a NodeId, a tuple of (str, str), or a sequence of PendingIdentifier objects."
            )

    def _set_pending_ids(self, identifiers: Sequence[PendingInstanceId]) -> ExtendedFileMetadataList:
        tasks = [
            {
                "url_path": f"{self._RESOURCE_PATH}/set-pending-instance-ids",
                "json": {
                    "items": [identifier.dump(camel_case=True) for identifier in id_chunk],
                },
                "api_subversion": "alpha",
            }
            for id_chunk in split_into_chunks(list(identifiers), self._PENDING_IDS_LIMIT)
        ]
        tasks_summary = execute_tasks(
            self._post,
            tasks,
            max_workers=self._config.max_workers,
            fail_fast=True,
        )
        tasks_summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=unpack_items_in_payload,
        )

        retrieved_items = tasks_summary.joined_results(lambda res: res.json()["items"])

        return ExtendedFileMetadataList._load(retrieved_items, cognite_client=self._cognite_client)

    def retrieve(
        self, id: int | None = None, external_id: str | None = None, instance_id: NodeId | None = None
    ) -> ExtendedFileMetadata | None:
        """`Retrieve a single file metadata by id. <https://developer.cognite.com/api#tag/Files/operation/getFileByInternalId>`_

        Args:
            id (int | None): ID
            external_id (str | None): External ID
            instance_id (NodeId | None): Instance ID

        Returns:
            FileMetadata | None: Requested file metadata or None if it does not exist.

        Examples:

            Get file metadata by id:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.files.retrieve(id=1)

            Get file metadata by external id:

                >>> res = client.files.retrieve(external_id="1")
        """
        identifiers = IdentifierSequence.load(ids=id, external_ids=external_id, instance_ids=instance_id).as_singleton()
        return self._retrieve_multiple(
            list_cls=ExtendedFileMetadataList,
            resource_cls=ExtendedFileMetadata,
            identifiers=identifiers,
            api_subversion="alpha",
        )

    def retrieve_multiple(
        self,
        ids: Sequence[int] | None = None,
        external_ids: SequenceNotStr[str] | None = None,
        instance_ids: Sequence[NodeId] | None = None,
        ignore_unknown_ids: bool = False,
    ) -> ExtendedFileMetadataList:
        """`Retrieve multiple file metadatas by id. <https://developer.cognite.com/api#tag/Files/operation/byIdsFiles>`_

        Args:
            ids (Sequence[int] | None): IDs
            external_ids (SequenceNotStr[str] | None): External IDs
            instance_ids (Sequence[NodeId] | None): Instance IDs
            ignore_unknown_ids (bool): Ignore IDs and external IDs that are not found rather than throw an exception.

        Returns:
            FileMetadataList: The requested file metadatas.

        Examples:

            Get file metadatas by id:

                >>> from cognite.client import CogniteClient
                >>> client = CogniteClient()
                >>> res = client.files.retrieve_multiple(ids=[1, 2, 3])

            Get file_metadatas by external id:

                >>> res = client.files.retrieve_multiple(external_ids=["abc", "def"])
        """
        identifiers = IdentifierSequence.load(ids=ids, external_ids=external_ids, instance_ids=instance_ids)
        return self._retrieve_multiple(
            list_cls=ExtendedFileMetadataList,
            resource_cls=ExtendedFileMetadata,
            identifiers=identifiers,
            ignore_unknown_ids=ignore_unknown_ids,
            api_subversion="alpha",
        )

    @overload
    def unlink_instance_ids(
        self,
        id: int | None = None,
        external_id: str | None = None,
    ) -> ExtendedFileMetadata | None: ...

    @overload
    def unlink_instance_ids(
        self,
        id: Sequence[int] | None = None,
        external_id: SequenceNotStr[str] | None = None,
    ) -> ExtendedFileMetadataList: ...

    def unlink_instance_ids(
        self,
        id: int | Sequence[int] | None = None,
        external_id: str | SequenceNotStr[str] | None = None,
    ) -> ExtendedFileMetadata | ExtendedFileMetadataList | None:
        """Unlink instance IDs from files.

        This allows a CogniteFile node in Data Modeling to be deleted without deleting the underlying file content.

        Args:
            id (int | Sequence[int] | None): The ID(s) of the files.
            external_id (str | SequenceNotStr[str] | None): The external ID(s) of the files.

        Returns:
            ExtendedFileMetadata | ExtendedFileMetadataList | None: The updated file metadata object(s). For single item requests, returns `None` if the file is not found.
        """
        if id is None and external_id is None:
            raise ValueError("At least one of id or external_id must be provided.")
        if isinstance(id, int) and isinstance(external_id, str):
            raise ValueError("Cannot specify both id and external_id as single values. Use one or the other.")
        is_list = isinstance(id, Sequence) or (isinstance(external_id, Sequence) and not isinstance(external_id, str))
        identifiers = IdentifierSequence.load(id, external_id)

        tasks = [
            {
                "url_path": f"{self._RESOURCE_PATH}/unlink-instance-ids",
                "json": {"items": id_chunk},
                "api_subversion": "alpha",
            }
            for id_chunk in split_into_chunks(identifiers.as_dicts(), 1000)
        ]
        tasks_summary = execute_tasks(
            self._post,
            tasks,
            max_workers=self._config.max_workers,
            fail_fast=True,
        )
        tasks_summary.raise_compound_exception_if_failed_tasks(
            task_unwrap_fn=unpack_items_in_payload,
        )

        retrieved_items = tasks_summary.joined_results(lambda res: res.json()["items"])

        result = ExtendedFileMetadataList._load(retrieved_items, cognite_client=self._cognite_client)
        if is_list:
            return result
        if len(result) == 0:
            return None
        if len(result) > 1:
            raise ValueError("Expected a single file, but multiple were returned.")
        return result[0]  # type: ignore[return-value]



import collections.abc
import io
from collections.abc import Hashable, Iterable
from pathlib import Path
from typing import Any, final

import pandas as pd
from cognite.client.data_classes import (
    AggregateResultItem,
    Asset,
    AssetList,
    AssetWrite,
    AssetWriteList,
    Event,
    EventList,
    EventWrite,
    EventWriteList,
    Sequence,
    SequenceList,
    SequenceWrite,
    SequenceWriteList,
    capabilities,
)
from cognite.client.data_classes.capabilities import Capability
from cognite.client.exceptions import CogniteAPIError, CogniteNotFoundError
from cognite.client.utils.useful_types import SequenceNotStr
from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.sequences import (
    ToolkitSequenceRows,
    ToolkitSequenceRowsList,
    ToolkitSequenceRowsWrite,
    ToolkitSequenceRowsWriteList,
)
from cognite_toolkit._cdf_tk.cruds._base_cruds import ResourceCRUD
from cognite_toolkit._cdf_tk.resource_classes import AssetYAML, EventYAML, SequenceRowYAML, SequenceYAML
from cognite_toolkit._cdf_tk.tk_warnings import LowSeverityWarning
from cognite_toolkit._cdf_tk.utils import load_yaml_inject_variables
from cognite_toolkit._cdf_tk.utils.diff_list import diff_list_hashable, diff_list_identifiable
from cognite_toolkit._cdf_tk.utils.file import read_csv

from .data_organization import DataSetsCRUD, LabelCRUD


@final
class AssetCRUD(ResourceCRUD[str, AssetWrite, Asset, AssetWriteList, AssetList]):
    folder_name = "classic"
    filename_pattern = r"^.*\.Asset$"  # Matches all yaml files whose stem ends with '.Asset'.
    filetypes = frozenset({"yaml", "yml", "csv", "parquet"})
    resource_cls = Asset
    resource_write_cls = AssetWrite
    list_cls = AssetList
    list_write_cls = AssetWriteList
    yaml_cls = AssetYAML
    kind = "Asset"
    dependencies = frozenset({DataSetsCRUD, LabelCRUD})
    _doc_url = "Assets/operation/createAssets"

    @property
    def display_name(self) -> str:
        return "assets"

    @classmethod
    def get_id(cls, item: Asset | AssetWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("Asset must have external_id")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: Asset | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("Asset must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[AssetWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: capabilities.AssetsAcl.Scope.All | capabilities.AssetsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            capabilities.AssetsAcl.Scope.All()
        )

        actions = (
            [capabilities.AssetsAcl.Action.Read]
            if read_only
            else [capabilities.AssetsAcl.Action.Read, capabilities.AssetsAcl.Action.Write]
        )

        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.AssetsAcl.Scope.DataSet(list(data_set_ids))

        return capabilities.AssetsAcl(actions, scope)

    def create(self, items: AssetWriteList) -> AssetList:
        return self.client.assets.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> AssetList:
        return self.client.assets.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: AssetWriteList) -> AssetList:
        return self.client.assets.update(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str | int]) -> int:
        if not ids:
            return 0
        internal_ids, external_ids = self._split_ids(ids)
        try:
            self.client.assets.delete(id=internal_ids, external_id=external_ids)
        except CogniteNotFoundError as e:
            # Do a CogniteNotFoundError instead of passing 'ignore_unknown_ids=True' to the delete method
            # to obtain an accurate list of deleted assets.
            non_existing = set(e.failed or [])
            if existing := [id_ for id_ in ids if id_ not in non_existing]:
                internal_ids, external_ids = self._split_ids(existing)
                self.client.assets.delete(id=internal_ids, external_id=external_ids)
            return len(existing)
        else:
            return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Asset]:
        return iter(
            self.client.assets(
                data_set_external_ids=[data_set_external_id] if data_set_external_id else None,
                # This is used in the purge command to delete the children before the parent.
                aggregated_properties=["depth", "child_count", "path"],
            )
        )

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]
        for label in item.get("labels", []):
            if isinstance(label, dict):
                yield LabelCRUD, label["externalId"]
            elif isinstance(label, str):
                yield LabelCRUD, label
        if "parentExternalId" in item:
            yield cls, item["parentExternalId"]

    def load_resource_file(
        self, filepath: Path, environment_variables: dict[str, str | None] | None = None
    ) -> list[dict[str, Any]]:
        resources: list[dict[str, Any]]
        if filepath.suffix in {".yaml", ".yml"}:
            raw_yaml = load_yaml_inject_variables(
                self.safe_read(filepath),
                environment_variables or {},
                original_filepath=filepath,
            )
            resources = [raw_yaml] if isinstance(raw_yaml, dict) else raw_yaml
        elif filepath.suffix == ".csv" or filepath.suffix == ".parquet":
            if filepath.suffix == ".csv":
                # The replacement is used to ensure that we read exactly the same file on Windows and Linux
                file_content = filepath.read_bytes().replace(b"\r\n", b"\n").decode("utf-8")
                data = read_csv(io.StringIO(file_content))
            else:
                data = pd.read_parquet(filepath)
            data.replace(pd.NA, None, inplace=True)
            data.replace("", None, inplace=True)
            resources = data.to_dict(orient="records")
        else:
            raise ValueError(f"Unsupported file type: {filepath.suffix}")

        return resources

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> AssetWrite:
        # Unpack metadata keys from table formats (e.g. csv, parquet)
        metadata: dict = resource.get("metadata", {})
        for key, value in list(resource.items()):
            if key.startswith("metadata."):
                if value not in {None, float("nan")} and str(value) not in {"", " ", "nan", "null", "none"}:
                    metadata[key.removeprefix("metadata.")] = str(value)
                del resource[key]
        if metadata:
            resource["metadata"] = metadata
        if isinstance(resource.get("labels"), str):
            resource["labels"] = [
                label.strip() for label in resource["labels"].removeprefix("[").removesuffix("]").split(",")
            ]

        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        return AssetWrite._load(resource)

    def dump_resource(self, resource: Asset, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        if "parentId" in dumped and "parentId" not in local:
            dumped.pop("parentId")
        if resource.aggregates:
            # This is only included when the asset is downloaded/migrated with aggregated properties
            aggregates = (
                resource.aggregates.dump()
                if isinstance(resource.aggregates, AggregateResultItem)
                else resource.aggregates
            )
            if "path" in aggregates:
                path = aggregates.pop("path", [])
                if path:
                    aggregates["path"] = self.client.lookup.assets.external_id(
                        [segment["id"] for segment in path if "id" in segment]
                    )
            dumped.update(aggregates)
        return dumped


@final
class SequenceCRUD(ResourceCRUD[str, SequenceWrite, Sequence, SequenceWriteList, SequenceList]):
    folder_name = "classic"
    filename_pattern = r"^.*\.Sequence$"
    resource_cls = Sequence
    resource_write_cls = SequenceWrite
    list_cls = SequenceList
    list_write_cls = SequenceWriteList
    kind = "Sequence"
    dependencies = frozenset({DataSetsCRUD, AssetCRUD})
    yaml_cls = SequenceYAML
    _doc_url = "Sequences/operation/createSequence"

    @property
    def display_name(self) -> str:
        return "sequences"

    @classmethod
    def get_id(cls, item: Sequence | SequenceWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("Sequence must have external_id")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: Sequence | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("Sequence must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[SequenceWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: Any = capabilities.SequencesAcl.Scope.All()
        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.SequencesAcl.Scope.DataSet(list(data_set_ids))

        actions = (
            [capabilities.SequencesAcl.Action.Read]
            if read_only
            else [capabilities.SequencesAcl.Action.Read, capabilities.SequencesAcl.Action.Write]
        )

        return capabilities.SequencesAcl(actions, scope)

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> SequenceWrite:
        if ds_external_id := resource.pop("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if asset_external_id := resource.pop("assetExternalId", None):
            resource["assetId"] = self.client.lookup.assets.id(asset_external_id, is_dry_run)
        return SequenceWrite._load(resource)

    def dump_resource(self, resource: Sequence, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if asset_id := dumped.pop("assetId", None):
            dumped["assetExternalId"] = self.client.lookup.assets.external_id(asset_id)
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        local_col_by_id = {col["externalId"]: col for col in local.get("columns", []) if "externalId" in col}
        for col in dumped.get("columns", []):
            external_id = col.get("externalId")
            if not external_id:
                continue
            if external_id not in local_col_by_id:
                continue
            local_col = local_col_by_id[external_id]
            if not col.get("metadata") and "metadata" not in local_col:
                col.pop("metadata", None)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path != ("columns",):
            return super().diff_list(local, cdf, json_path)
        return diff_list_identifiable(local, cdf, get_identifier=lambda col: col["externalId"])

    def create(self, items: SequenceWriteList) -> SequenceList:
        return self.client.sequences.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> SequenceList:
        return self.client.sequences.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: SequenceWriteList) -> SequenceList:
        return self.client.sequences.upsert(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str | int]) -> int:
        internal_ids, external_ids = self._split_ids(ids)
        try:
            self.client.sequences.delete(id=internal_ids, external_id=external_ids)
        except (CogniteAPIError, CogniteNotFoundError) as e:
            non_existing = set(e.failed or [])
            if existing := [id_ for id_ in ids if id_ not in non_existing]:
                internal_ids, external_ids = self._split_ids(existing)
                self.client.sequences.delete(id=internal_ids, external_id=external_ids)
            return len(existing)
        else:
            return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Sequence]:
        return iter(
            self.client.sequences(data_set_external_ids=[data_set_external_id] if data_set_external_id else None)
        )

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]
        if "assetExternalId" in item:
            yield AssetCRUD, item["assetExternalId"]


@final
class SequenceRowCRUD(
    ResourceCRUD[
        str, ToolkitSequenceRowsWrite, ToolkitSequenceRows, ToolkitSequenceRowsWriteList, ToolkitSequenceRowsList
    ]
):
    folder_name = "classic"
    filename_pattern = r"^.*\.SequenceRow$"
    resource_cls = ToolkitSequenceRows
    resource_write_cls = ToolkitSequenceRowsWrite
    list_cls = ToolkitSequenceRowsList
    list_write_cls = ToolkitSequenceRowsWriteList
    kind = "SequenceRow"
    dependencies = frozenset({SequenceCRUD})
    parent_resource = frozenset({SequenceCRUD})
    _doc_url = "Sequences/operation/postSequenceData"
    yaml_cls = SequenceRowYAML
    support_update = False

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None):
        super().__init__(client, build_dir, console)
        # Used in the .diff_list method to keep track of the last column in the local list
        # such that the values in the rows can be matched to the correct column.
        self._last_column: tuple[dict[int, int], list[int]] = {}, []

    @property
    def display_name(self) -> str:
        return "sequence rows"

    @classmethod
    def get_id(cls, item: ToolkitSequenceRows | ToolkitSequenceRowsWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("SequenceRows must have external_id")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: ToolkitSequenceRows | ToolkitSequenceRowsWrite | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("SequenceRows must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[ToolkitSequenceRowsWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        # We don't have any capabilities for SequenceRows, that is already handled by the Sequence
        return []

    def create(self, items: ToolkitSequenceRowsWriteList) -> ToolkitSequenceRowsWriteList:
        item: ToolkitSequenceRowsWrite
        for item in items:
            self.client.sequences.rows.insert(item.as_sequence_rows(), external_id=item.external_id)
        return items

    def retrieve(self, ids: SequenceNotStr[str]) -> ToolkitSequenceRowsList:
        retrieved = self.client.sequences.rows.retrieve(external_id=ids)
        return ToolkitSequenceRowsList([ToolkitSequenceRows._load(row.dump(camel_case=True)) for row in retrieved])

    def delete(self, ids: SequenceNotStr[str]) -> int:
        for id_ in ids:
            self.client.sequences.rows.delete_range(start=0, end=None, external_id=id_)
        return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[ToolkitSequenceRows]:
        if parent_ids is None:
            sequence_iterable = self.client.sequences(
                data_set_external_ids=[data_set_external_id] if data_set_external_id else None
            )
            parent_ids = [seq.external_id or seq.id for seq in sequence_iterable]
        for sequence_id in parent_ids:
            if isinstance(sequence_id, str):
                res = self.client.sequences.rows.retrieve(external_id=sequence_id)
                if res:
                    yield ToolkitSequenceRows._load(res.dump(camel_case=True))
            elif isinstance(sequence_id, int):
                res = self.client.sequences.rows.retrieve(id=sequence_id)
                if res:
                    yield ToolkitSequenceRows._load(res.dump(camel_case=True))

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        yield SequenceCRUD, item["externalId"]

    def dump_resource(self, resource: ToolkitSequenceRows, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()

        if local is not None and "id" in dumped and "id" not in local:
            dumped.pop("id")
        # Ensure that the rows is the last key in the dumped dictionary,
        # This information is used in the .diff_list method to match the values in the rows to the correct column.
        dumped["rows"] = dumped.pop("rows")
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path == ("rows",):
            return diff_list_identifiable(local, cdf, get_identifier=lambda row: row["rowNumber"])
        elif json_path == ("columns",):
            self._last_column = diff_list_hashable(local, cdf)
            # This is a special case where we need to keep track of the last column in the local list
            # such that the values in the rows can be matched to the correct column.
            return self._last_column
        elif len(json_path) == 3 and json_path[0] == "rows" and json_path[2] == "values":
            local_by_cdf, added = self._last_column
            if len(cdf) == len(local_by_cdf) + len(added):
                return local_by_cdf, added
            else:
                LowSeverityWarning("Number of rows in does not match the number of columns").print_warning()
                # Just assume that the rows are in the correct order
                return {no: no for no in range(len(cdf))}, []
        return super().diff_list(local, cdf, json_path)


@final
class EventCRUD(ResourceCRUD[str, EventWrite, Event, EventWriteList, EventList]):
    folder_name = "classic"
    filename_pattern = r"^.*\.Event$"  # Matches all yaml files whose stem ends with '.Event'.
    filetypes = frozenset({"yaml", "yml"})
    resource_cls = Event
    resource_write_cls = EventWrite
    list_cls = EventList
    list_write_cls = EventWriteList
    yaml_cls = EventYAML
    kind = "Event"
    dependencies = frozenset({DataSetsCRUD, AssetCRUD})
    _doc_url = "Events/operation/createEvents"

    @property
    def display_name(self) -> str:
        return "events"

    @classmethod
    def get_id(cls, item: Event | EventWrite | dict) -> str:
        if isinstance(item, dict):
            return item["externalId"]
        if not item.external_id:
            raise KeyError("Event must have external_id")
        return item.external_id

    @classmethod
    def get_internal_id(cls, item: Event | dict) -> int:
        if isinstance(item, dict):
            return item["id"]
        if not item.id:
            raise KeyError("Event must have id")
        return item.id

    @classmethod
    def dump_id(cls, id: str) -> dict[str, Any]:
        return {"externalId": id}

    @classmethod
    def get_required_capability(
        cls, items: collections.abc.Sequence[EventWrite] | None, read_only: bool
    ) -> Capability | list[Capability]:
        if not items and items is not None:
            return []
        scope: capabilities.EventsAcl.Scope.All | capabilities.EventsAcl.Scope.DataSet = (  # type: ignore[valid-type]
            capabilities.EventsAcl.Scope.All()
        )

        actions = (
            [capabilities.EventsAcl.Action.Read]
            if read_only
            else [capabilities.EventsAcl.Action.Read, capabilities.EventsAcl.Action.Write]
        )

        if items:
            if data_set_ids := {item.data_set_id for item in items if item.data_set_id}:
                scope = capabilities.EventsAcl.Scope.DataSet(list(data_set_ids))

        return capabilities.EventsAcl(actions, scope)

    def create(self, items: EventWriteList) -> EventList:
        return self.client.events.create(items)

    def retrieve(self, ids: SequenceNotStr[str]) -> EventList:
        return self.client.events.retrieve_multiple(external_ids=ids, ignore_unknown_ids=True)

    def update(self, items: EventWriteList) -> EventList:
        return self.client.events.update(items, mode="replace")

    def delete(self, ids: SequenceNotStr[str | int]) -> int:
        internal_ids, external_ids = self._split_ids(ids)
        try:
            self.client.events.delete(id=internal_ids, external_id=external_ids)
        except (CogniteAPIError, CogniteNotFoundError) as e:
            non_existing = set(e.failed or [])
            if existing := [id_ for id_ in ids if id_ not in non_existing]:
                internal_ids, external_ids = self._split_ids(existing)
                self.client.events.delete(id=internal_ids, external_id=external_ids)
            return len(existing)
        else:
            return len(ids)

    def _iterate(
        self,
        data_set_external_id: str | None = None,
        space: str | None = None,
        parent_ids: list[Hashable] | None = None,
    ) -> Iterable[Event]:
        return iter(self.client.events(data_set_external_ids=[data_set_external_id] if data_set_external_id else None))

    @classmethod
    def get_dependent_items(cls, item: dict) -> Iterable[tuple[type[ResourceCRUD], Hashable]]:
        """Returns all items that this item requires.

        For example, a TimeSeries requires a DataSet, so this method would return the
        DatasetLoader and identifier of that dataset.
        """
        if "dataSetExternalId" in item:
            yield DataSetsCRUD, item["dataSetExternalId"]
        for asset_id in item.get("assetExternalIds", []):
            if isinstance(asset_id, str):
                yield AssetCRUD, asset_id

    def load_resource(self, resource: dict[str, Any], is_dry_run: bool = False) -> EventWrite:
        if ds_external_id := resource.get("dataSetExternalId", None):
            resource["dataSetId"] = self.client.lookup.data_sets.id(ds_external_id, is_dry_run)
        if asset_external_ids := resource.pop("assetExternalIds", []):
            resource["assetIds"] = self.client.lookup.assets.id(asset_external_ids, is_dry_run)
        return EventWrite._load(resource)

    def dump_resource(self, resource: Event, local: dict[str, Any] | None = None) -> dict[str, Any]:
        dumped = resource.as_write().dump()
        local = local or {}
        if data_set_id := dumped.pop("dataSetId", None):
            dumped["dataSetExternalId"] = self.client.lookup.data_sets.external_id(data_set_id)
        if asset_ids := dumped.pop("assetIds", None):
            local_order = {asset: no for no, asset in enumerate(local.get("assetExternalIds", []))}
            end_of_list = len(local_order)
            dumped["assetExternalIds"] = sorted(
                self.client.lookup.assets.external_id(asset_ids), key=lambda a: local_order.get(a, end_of_list)
            )
        if not dumped.get("metadata") and "metadata" not in local:
            dumped.pop("metadata", None)
        return dumped

    def diff_list(
        self, local: list[Any], cdf: list[Any], json_path: tuple[str | int, ...]
    ) -> tuple[dict[int, int], list[int]]:
        if json_path != ("assetExternalIds",):
            return super().diff_list(local, cdf, json_path)
        return diff_list_hashable(local, cdf)



from collections.abc import Callable, Iterable, Sequence
from enum import Enum
from pathlib import Path

from cognite.client.data_classes._base import T_CogniteResource
from rich import print
from rich.console import Console
from rich.table import Table

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.commands._base import ToolkitCommand
from cognite_toolkit._cdf_tk.commands._migrate.creators import MigrationCreator
from cognite_toolkit._cdf_tk.commands._migrate.data_mapper import DataMapper
from cognite_toolkit._cdf_tk.commands.deploy import DeployCommand
from cognite_toolkit._cdf_tk.constants import DMS_INSTANCE_LIMIT_MARGIN
from cognite_toolkit._cdf_tk.cruds import ResourceWorker
from cognite_toolkit._cdf_tk.data_classes import DeployResults
from cognite_toolkit._cdf_tk.exceptions import (
    ToolkitFileExistsError,
    ToolkitMigrationError,
    ToolkitValueError,
)
from cognite_toolkit._cdf_tk.storageio import T_Selector, UploadableStorageIO, UploadItem
from cognite_toolkit._cdf_tk.utils import humanize_collection, safe_write, sanitize_filename
from cognite_toolkit._cdf_tk.utils.file import yaml_safe_dump
from cognite_toolkit._cdf_tk.utils.fileio import Chunk, CSVWriter, NDJsonWriter, SchemaColumn, Uncompressed
from cognite_toolkit._cdf_tk.utils.http_client import HTTPClient, HTTPMessage, ItemMessage, SuccessResponseItems
from cognite_toolkit._cdf_tk.utils.producer_worker import ProducerWorkerExecutor
from cognite_toolkit._cdf_tk.utils.progress_tracker import AVAILABLE_STATUS, ProgressTracker, Status
from cognite_toolkit._cdf_tk.utils.useful_types import T_WriteCogniteResource

from .data_model import INSTANCE_SOURCE_VIEW_ID, MODEL_ID, RESOURCE_VIEW_MAPPING_VIEW_ID


class MigrationCommand(ToolkitCommand):
    class Steps(str, Enum):
        DOWNLOAD = "download"
        CONVERT = "convert"
        UPLOAD = "upload"

        @classmethod
        def list(cls) -> list[str]:
            return [step.value for step in cls.__members__.values()]

    def migrate(
        self,
        selected: T_Selector,
        data: UploadableStorageIO[T_Selector, T_CogniteResource, T_WriteCogniteResource],
        mapper: DataMapper[T_Selector, T_CogniteResource, T_WriteCogniteResource],
        log_dir: Path,
        dry_run: bool = False,
        verbose: bool = False,
    ) -> ProgressTracker[str]:
        if log_dir.exists() and any(log_dir.iterdir()):
            raise ToolkitFileExistsError(
                f"Log directory {log_dir} already exists. Please remove it or choose another directory."
            )
        self.validate_migration_model_available(data.client)
        log_dir.mkdir(parents=True, exist_ok=True)
        mapper.prepare(selected)

        iteration_count: int | None = None
        total_items = data.count(selected)
        if total_items is not None:
            iteration_count = (total_items // data.CHUNK_SIZE) + (1 if total_items % data.CHUNK_SIZE > 0 else 0)
            self.validate_available_capacity(data.client, total_items)

        console = Console()
        tracker = ProgressTracker[str](self.Steps.list())
        with (
            NDJsonWriter(log_dir, kind=f"{selected.kind}MigrationIssues", compression=Uncompressed) as log_file,
            HTTPClient(config=data.client.config) as write_client,
        ):
            executor = ProducerWorkerExecutor[
                Sequence[T_CogniteResource], Sequence[UploadItem[T_WriteCogniteResource]]
            ](
                download_iterable=self._download_iterable(selected, data, tracker),
                process=self._convert(mapper, data, tracker, log_file),
                write=self._upload(selected, write_client, data, tracker, log_file, dry_run),
                iteration_count=iteration_count,
                max_queue_size=10,
                download_description=f"Downloading {selected.display_name}",
                process_description="Converting",
                write_description="Uploading",
                console=console,
                verbose=verbose,
            )

            executor.run()
            total = executor.total_items

        self._print_table(tracker.aggregate(), console)
        self._print_csv(tracker, log_dir, f"{selected.kind}Items", console)
        executor.raise_on_error()
        action = "Would migrate" if dry_run else "Migrating"
        console.print(f"{action} {total:,} {selected.display_name} to instances.")
        return tracker

    def _print_table(self, results: dict[tuple[str, Status], int], console: Console) -> None:
        for step in self.Steps:
            # We treat pending as failed for summary purposes
            results[(step.value, "failed")] = results.get((step.value, "failed"), 0) + results.get(
                (step.value, "pending"), 0
            )

        table = Table(title="Migration Summary", show_lines=True)
        table.add_column("Status", style="cyan", no_wrap=True)
        for step in self.Steps:
            table.add_column(step.value.capitalize(), style="magenta")
        for status in AVAILABLE_STATUS:
            if status == "pending":
                # Skip pending as we treat it as failed
                continue
            row = [status]
            for step in self.Steps:
                row.append(str(results.get((step.value, status), 0)))
            table.add_row(*row)

        console.print(table)

    def _print_csv(self, tracker: ProgressTracker[str], log_dir: Path, kind: str, console: Console) -> None:
        with CSVWriter(log_dir, kind=kind, compression=Uncompressed, columns=self._csv_columns()) as csv_file:
            batch: list[Chunk] = []
            steps = self.Steps.list()
            for item_id, progress in tracker.result().items():
                batch.append({"ID": str(item_id), **{step: progress[step] for step in steps}})
                if len(batch) >= 1000:
                    csv_file.write_chunks(batch)
                    batch = []
            if batch:
                csv_file.write_chunks(batch)
        console.print(f"Migration items written to {log_dir}")

    @classmethod
    def _csv_columns(cls) -> list[SchemaColumn]:
        return [
            SchemaColumn(name="ID", type="string"),
            *(SchemaColumn(name=step, type="string") for step in cls.Steps.list()),
        ]

    def _download_iterable(
        self,
        selected: T_Selector,
        data: UploadableStorageIO[T_Selector, T_CogniteResource, T_WriteCogniteResource],
        tracker: ProgressTracker[str],
    ) -> Iterable[Sequence[T_CogniteResource]]:
        for page in data.stream_data(selected):
            for item in page.items:
                tracker.set_progress(data.as_id(item), self.Steps.DOWNLOAD, "success")
            yield page.items

    def _convert(
        self,
        mapper: DataMapper[T_Selector, T_CogniteResource, T_WriteCogniteResource],
        data: UploadableStorageIO[T_Selector, T_CogniteResource, T_WriteCogniteResource],
        tracker: ProgressTracker[str],
        log_file: NDJsonWriter,
    ) -> Callable[[Sequence[T_CogniteResource]], Sequence[UploadItem[T_WriteCogniteResource]]]:
        def track_mapping(source: Sequence[T_CogniteResource]) -> list[UploadItem[T_WriteCogniteResource]]:
            issues: list[Chunk] = []
            targets: list[UploadItem[T_WriteCogniteResource]] = []
            for item in source:
                target, issue = mapper.map(item)
                id_ = data.as_id(item)
                result: Status = "failed" if target is None else "success"
                tracker.set_progress(id_, step=self.Steps.CONVERT, status=result)

                if issue.has_issues:
                    # MyPy fails to understand that dict[str, JsonVal] is a Chunk
                    issues.append(issue.dump())  # type: ignore[arg-type]
                if target is not None:
                    targets.append(UploadItem(source_id=id_, item=target))
            if issues:
                log_file.write_chunks(issues)
            return targets

        return track_mapping

    def _upload(
        self,
        selected: T_Selector,
        write_client: HTTPClient,
        target: UploadableStorageIO[T_Selector, T_CogniteResource, T_WriteCogniteResource],
        tracker: ProgressTracker[str],
        log_file: NDJsonWriter,
        dry_run: bool,
    ) -> Callable[[Sequence[UploadItem[T_WriteCogniteResource]]], None]:
        def upload_items(data_item: Sequence[UploadItem[T_WriteCogniteResource]]) -> None:
            if not data_item:
                return None
            responses: Sequence[HTTPMessage]
            if dry_run:
                responses = [SuccessResponseItems(200, "", [item.source_id for item in data_item])]
            else:
                responses = target.upload_items(data_chunk=data_item, http_client=write_client, selector=selected)

            issues: list[Chunk] = []
            for item in responses:
                if isinstance(item, SuccessResponseItems):
                    for success_id in item.ids:
                        tracker.set_progress(success_id, step=self.Steps.UPLOAD, status="success")
                elif isinstance(item, ItemMessage):
                    for failed_id in item.ids:
                        tracker.set_progress(failed_id, step=self.Steps.UPLOAD, status="failed")

                if not isinstance(item, SuccessResponseItems):
                    issues.append(item.dump())  # type: ignore[arg-type]
            if issues:
                log_file.write_chunks(issues)
            return None

        return upload_items

    @staticmethod
    def validate_migration_model_available(client: ToolkitClient) -> None:
        models = client.data_modeling.data_models.retrieve([MODEL_ID], inline_views=False)
        if not models:
            raise ToolkitMigrationError(
                f"The migration data model {MODEL_ID!r} does not exist. "
                "Please run the `cdf migrate prepare` command to deploy the migration data model."
            )
        elif len(models) > 1:
            raise ToolkitMigrationError(
                f"Multiple migration models {MODEL_ID!r}. "
                "Please delete the duplicate models before proceeding with the migration."
            )
        model = models[0]
        missing_views = {INSTANCE_SOURCE_VIEW_ID, RESOURCE_VIEW_MAPPING_VIEW_ID} - set(model.views or [])
        if missing_views:
            raise ToolkitMigrationError(
                f"Invalid migration model. Missing views {humanize_collection(missing_views)}. "
                f"Please run the `cdf migrate prepare` command to deploy the migration data model."
            )

    def validate_available_capacity(self, client: ToolkitClient, instance_count: int) -> None:
        """Validate that the project has enough capacity to accommodate the migration."""

        stats = client.data_modeling.statistics.project()

        available_capacity = stats.instances.instances_limit - stats.instances.instances
        available_capacity_after = available_capacity - instance_count

        if available_capacity_after < DMS_INSTANCE_LIMIT_MARGIN:
            raise ToolkitValueError(
                "Cannot proceed with migration, not enough instance capacity available. Total capacity after migration"
                f" would be {available_capacity_after:,} instances, which is less than the required margin of"
                f" {DMS_INSTANCE_LIMIT_MARGIN:,} instances. Please increase the instance capacity in your CDF project"
                f" or delete some existing instances before proceeding with the migration of {instance_count:,} assets."
            )
        total_instances = stats.instances.instances + instance_count
        self.console(
            f"Project has enough capacity for migration. Total instances after migration: {total_instances:,}."
        )

    def create(
        self,
        client: ToolkitClient,
        creator: MigrationCreator,
        dry_run: bool,
        output_dir: Path,
        verbose: bool = False,
    ) -> DeployResults:
        """This method is used to create migration resource in CDF."""
        self.validate_migration_model_available(client)

        deploy_cmd = DeployCommand(self.print_warning, silent=self.silent)
        deploy_cmd.tracker = self.tracker

        crud_cls = creator.CRUD
        resource_list = creator.create_resources()

        results = DeployResults([], "deploy", dry_run=dry_run)
        crud = crud_cls.create_loader(client)
        worker = ResourceWorker(crud, "deploy")
        local_by_id = {crud.get_id(item): (item.dump(), item) for item in resource_list}
        worker.validate_access(local_by_id, is_dry_run=dry_run)
        cdf_resources = crud.retrieve(list(local_by_id.keys()))
        resources = worker.categorize_resources(local_by_id, cdf_resources, False, verbose)

        if dry_run:
            result = deploy_cmd.dry_run_deploy(resources, crud, False, False)
        else:
            result = deploy_cmd.actual_deploy(resources, crud)
            if result.calculated_total > 0 and creator.HAS_LINEAGE:
                store_count = creator.store_lineage(resource_list)
                self.console(f"Stored lineage for {store_count:,} {creator.DISPLAY_NAME}.")

        verb = "Would deploy" if dry_run else "Deploying"
        self.console(f"{verb} {creator.DISPLAY_NAME} to CDF.")

        resource_configs = creator.resource_configs(resource_list)
        for config in resource_configs:
            filepath = output_dir / crud_cls.folder_name / f"{sanitize_filename(config.filestem)}.{crud_cls.kind}.yaml"
            filepath.parent.mkdir(parents=True, exist_ok=True)
            safe_write(filepath, yaml_safe_dump(config.data))
        self.console(
            f"{len(resource_configs)} {crud_cls.kind} resource configurations written to {(output_dir / crud_cls.folder_name).as_posix()!r}"
        )
        self.console("It is recommended to add these files to a Toolkit governed module.")

        if result:
            results[result.name] = result

        if results.has_counts:
            print(results.counts_table())

        return results
